# -*- coding: utf-8 -*-
import json
import logging
import os
import re
import threading
import time
import queue
import ssl
import urllib.parse
import urllib.request
from typing import Dict, Optional, Any

try:
    from flask import Flask, request, jsonify, Response, send_from_directory
except ImportError:
    raise ImportError(
        "Flask is required for web server. "
        "Install it with: pip install acp-py[web]"
    ) from None

from agentcp.agentcp import AgentCP, AgentID
from agentcp.base.log import log_info, log_error, log_warning
from agentcp.utils import tls_config as tls
from agentcp.group_client import GroupClient, GroupOperations

app = Flask(__name__)
app.logger.setLevel(logging.WARNING)
logging.getLogger('werkzeug').setLevel(logging.WARNING)

# ============================================================
# Global State
# ============================================================
MAX_AIDS = 10
global_data_dir = ''
global_seed_password = ''
acp: Optional[AgentCP] = None

class AidInstance:
    def __init__(self, aid: str, agent_id: AgentID):
        self.aid = aid
        self.agent_id = agent_id
        self.online = False
        self.ws_connected = False
        self.ws_status = 'disconnected'
        self.active_group_id = None
        self.group_initialized = False
        self.group_client: Optional[GroupClient] = None
        self.group_ops: Optional[GroupOperations] = None
        self.group_target_aid = ''
        self.group_session_id = ''

aid_instances: Dict[str, AidInstance] = {}

# P2P message store: {aid: {session_id: {messages:[], peer_aid:str, type:str, closed:bool}}}
message_stores: Dict[str, dict] = {}

# Agent info cache
agent_info_cache: Dict[str, dict] = {}
AGENT_INFO_CACHE_TTL = 24 * 60 * 60

# Group message store: {group_id: [messages]}
group_message_stores: Dict[str, list] = {}
group_message_lock = threading.Lock()
# Group list store: {aid: [{group_id, name}]}
group_list_stores: Dict[str, list] = {}

# AID md options (nickname, description)
aid_md_options: Dict[str, dict] = {}

# Group message polling threads: {aid: threading.Thread}
group_poll_threads: Dict[str, threading.Thread] = {}
group_poll_running: Dict[str, bool] = {}
# Per-AID poll cursor: {aid: {group_id: last_pushed_msg_id}}
group_poll_cursors: Dict[str, Dict[str, int]] = {}


def get_acp() -> AgentCP:
    global acp
    if acp is None:
        acp = AgentCP(global_data_dir, seed_password=global_seed_password, run_proxy=False)
    return acp


# ============================================================
# Message Store helpers (with file persistence)
# ============================================================
def _sessions_dir(aid: str) -> str:
    d = global_data_dir or os.getcwd()
    return os.path.join(d, 'agentcp', 'AIDs', aid, 'sessions')


def _flush_index(aid: str):
    """Save session index to _index.json"""
    try:
        sdir = _sessions_dir(aid)
        os.makedirs(sdir, exist_ok=True)
        store = message_stores.get(aid, {})
        records = []
        for sid, s in store.items():
            records.append({
                'session_id': sid, 'peer_aid': s['peer_aid'], 'type': s['type'],
                'identifying_code': s.get('identifying_code', ''),
                'last_message_at': s.get('last_message_at', 0),
                'message_count': len(s.get('messages', [])),
                'closed': s.get('closed', False),
            })
        with open(os.path.join(sdir, '_index.json'), 'w', encoding='utf-8') as f:
            json.dump(records, f, ensure_ascii=False, indent=2)
    except Exception:
        pass


def _append_message(aid: str, session_id: str, msg: dict):
    """Append one message to {session_id}.jsonl"""
    try:
        sdir = _sessions_dir(aid)
        os.makedirs(sdir, exist_ok=True)
        with open(os.path.join(sdir, f'{session_id}.jsonl'), 'a', encoding='utf-8') as f:
            f.write(json.dumps(msg, ensure_ascii=False) + '\n')
    except Exception:
        pass


def _rename_session_file(aid: str, old_sid: str, new_sid: str):
    """Rename session jsonl file from old_sid to new_sid."""
    try:
        sdir = _sessions_dir(aid)
        old_path = os.path.join(sdir, f'{old_sid}.jsonl')
        new_path = os.path.join(sdir, f'{new_sid}.jsonl')
        if os.path.exists(old_path):
            os.replace(old_path, new_path)
    except Exception:
        pass


def load_sessions_for_aid(aid: str):
    """Load sessions + messages from disk into memory"""
    sdir = _sessions_dir(aid)
    index_path = os.path.join(sdir, '_index.json')
    if not os.path.exists(index_path):
        return
    try:
        with open(index_path, 'r', encoding='utf-8') as f:
            records = json.load(f)
        if not isinstance(records, list):
            return
        store = get_store(aid)
        for r in records:
            sid = r.get('session_id', '')
            if not sid or sid in store:
                continue
            # Load messages from jsonl
            msgs = []
            msg_path = os.path.join(sdir, f'{sid}.jsonl')
            if os.path.exists(msg_path):
                with open(msg_path, 'r', encoding='utf-8') as f:
                    for line in f:
                        line = line.strip()
                        if line:
                            try:
                                msgs.append(json.loads(line))
                            except Exception:
                                pass
            store[sid] = {
                'session_id': sid,
                'peer_aid': r.get('peer_aid', ''),
                'type': r.get('type', 'incoming'),
                'identifying_code': r.get('identifying_code', ''),
                'messages': msgs,
                'closed': r.get('closed', False),
                'last_message_at': r.get('last_message_at', 0),
                'last_message': msgs[-1].get('content', '')[:50] if msgs else '',
            }
    except Exception:
        pass


def get_store(aid: str) -> dict:
    if aid not in message_stores:
        message_stores[aid] = {}
    return message_stores[aid]


def get_or_create_session(aid: str, session_id: str, peer_aid: str, stype: str, identifying_code: str = ''):
    store = get_store(aid)
    if session_id not in store:
        store[session_id] = {
            'session_id': session_id,
            'peer_aid': peer_aid,
            'type': stype,
            'identifying_code': identifying_code,
            'messages': [],
            'closed': False,
            'created_at': int(time.time() * 1000),
            'last_message_at': int(time.time() * 1000),
            'last_message': '',
        }
        _flush_index(aid)
    return store[session_id]


def add_message_to_session(aid: str, session_id: str, msg: dict):
    store = get_store(aid)
    if session_id in store:
        store[session_id]['messages'].append(msg)
        store[session_id]['last_message_at'] = msg.get('timestamp', int(time.time() * 1000))
        store[session_id]['last_message'] = msg.get('content', '')[:50]
        _append_message(aid, session_id, msg)
        _flush_index(aid)


def get_session_list(aid: str):
    store = get_store(aid)
    result = []
    for sid, s in store.items():
        result.append({
            'sessionId': sid,
            'peerAid': s['peer_aid'],
            'type': s['type'],
            'messageCount': len(s['messages']),
            'lastMessageAt': s.get('last_message_at', 0),
            'lastMessage': s.get('last_message', ''),
            'closed': s.get('closed', False),
        })
    result.sort(key=lambda x: x['lastMessageAt'], reverse=True)
    return result


# ============================================================
# Standalone WebSocket server (sync, thread-per-connection)
# ============================================================
from websockets.sync.server import serve as ws_serve
from websockets.exceptions import ConnectionClosed

ws_queues: list = []  # list of queue.Queue, one per browser connection
ws_lock = threading.Lock()


def push_to_browser(data: dict):
    msg = json.dumps(data)
    with ws_lock:
        for q in ws_queues:
            try:
                q.put_nowait(msg)
            except Exception:
                pass


def push_to_aid(aid: str, data: dict):
    # TODO: filter by aid when multi-browser support is added
    push_to_browser(data)


# ============================================================
# Agent info fetching
# ============================================================
def fetch_agent_md(aid: str) -> str:
    try:
        ctx = tls.ssl_context()
        req = urllib.request.Request(f'https://{aid}/agent.md', method='GET')
        with urllib.request.urlopen(req, timeout=5, context=ctx) as resp:
            return resp.read().decode('utf-8')
    except Exception:
        return ''


def parse_agent_md_frontmatter(content: str) -> dict:
    result = {'type': '', 'name': '', 'description': '', 'tags': []}
    m = re.match(r'^---\s*\n([\s\S]*?)\n---', content)
    if not m:
        return result
    yaml = m.group(1)
    for key in ['type', 'name', 'description']:
        km = re.search(rf'^{key}:\s*"?([^"\n]*)"?\s*$', yaml, re.MULTILINE)
        if km:
            result[key] = km.group(1).strip()
    tags_block = re.search(r'^tags:\s*\n((?:\s+-\s+.*\n?)*)', yaml, re.MULTILINE)
    if tags_block:
        tag_lines = re.findall(r'^\s+-\s+(.+)$', tags_block.group(1), re.MULTILINE)
        result['tags'] = [t.strip().strip('"') for t in tag_lines]
    return result


def get_agent_info(aid: str) -> dict:
    cached = agent_info_cache.get(aid)
    if cached and time.time() - cached.get('cached_at', 0) < AGENT_INFO_CACHE_TTL:
        return {k: cached[k] for k in ['type', 'name', 'description', 'tags']}
    try:
        md = fetch_agent_md(aid)
        info = parse_agent_md_frontmatter(md)
        agent_info_cache[aid] = {**info, 'cached_at': time.time()}
        return info
    except Exception:
        if cached:
            return {k: cached[k] for k in ['type', 'name', 'description', 'tags']}
        return {'type': '', 'name': '', 'description': '', 'tags': []}


# ============================================================
# AID md options persistence
# ============================================================
def get_aid_md_options_path() -> str:
    d = global_data_dir or os.getcwd()
    return os.path.join(d, 'agentcp', 'AIDs', '.aid-md-options.json')


def load_aid_md_options():
    global aid_md_options
    try:
        p = get_aid_md_options_path()
        if os.path.exists(p):
            with open(p, 'r', encoding='utf-8') as f:
                aid_md_options = json.load(f)
    except Exception:
        pass


def save_aid_md_options_for(aid: str, opts: dict):
    global aid_md_options
    aid_md_options[aid] = opts
    try:
        p = get_aid_md_options_path()
        os.makedirs(os.path.dirname(p), exist_ok=True)
        with open(p, 'w', encoding='utf-8') as f:
            json.dump(aid_md_options, f, ensure_ascii=False, indent=2)
    except Exception:
        pass


# ============================================================
# AID validation
# ============================================================
def validate_aid(aid: str) -> dict:
    try:
        cp = get_acp()
        cert_pem = cp.read_certificate_pem(aid)
        keys_exist = cert_pem is not None and len(cert_pem) > 0
        # Simple validity check
        cert_valid = keys_exist
        if keys_exist:
            try:
                from cryptography import x509
                from cryptography.hazmat.backends import default_backend
                import datetime
                cert = x509.load_pem_x509_certificate(cert_pem.encode(), default_backend())
                cert_valid = cert.not_valid_after_utc > datetime.datetime.now(datetime.timezone.utc)
            except Exception:
                cert_valid = keys_exist
        return {'keysExist': keys_exist, 'certValid': cert_valid}
    except Exception:
        return {'keysExist': False, 'certValid': False}


def get_aid_status_list() -> list:
    cp = get_acp()
    aid_list = cp.get_aid_list()
    result = []
    for aid in aid_list:
        v = validate_aid(aid)
        inst = aid_instances.get(aid)
        result.append({
            'aid': aid,
            'keysExist': v['keysExist'],
            'certValid': v['certValid'],
            'online': inst.online if inst else False,
        })
    return result


# ============================================================
# Ensure AID online
# ============================================================
online_locks: Dict[str, threading.Lock] = {}


def ensure_online(target_aid: str) -> AidInstance:
    if target_aid in aid_instances:
        inst = aid_instances[target_aid]
        if inst.online:
            return inst

    lock = online_locks.setdefault(target_aid, threading.Lock())
    with lock:
        if target_aid in aid_instances and aid_instances[target_aid].online:
            return aid_instances[target_aid]
        return do_ensure_online(target_aid)


def do_ensure_online(aid: str) -> AidInstance:
    load_sessions_for_aid(aid)
    cp = get_acp()
    agent_id = cp.load_aid(aid)
    if agent_id is None:
        raise Exception(f'AID {aid} 加载失败')

    # Set up message handler for P2P
    def on_message(data):
        handle_incoming_message(aid, data)

    agent_id.add_message_handler(on_message)

    # 在 online() 之前设置 agent.md options，这样 online() 内部的后台线程会使用正确的选项
    options = aid_md_options.get(aid, {})
    if options:
        agent_id.set_agent_md_options(options)

    agent_id.online()
    # online() 内部已通过后台线程自动处理 agent.md 上传，无需额外调用

    # Set up heartbeat invite handler
    if agent_id.heartbeat_client:
        original_on_invite = getattr(agent_id.heartbeat_client, 'on_recv_invite', None)
        def on_invite(invite_req):
            # invite_req is InviteMessageReq object, not dict
            session_id = getattr(invite_req, 'SessionId', '') or ''
            inviter = getattr(invite_req, 'InviterAgentId', '') or ''
            invite_code = getattr(invite_req, 'InviteCode', '') or ''
            if session_id and inviter:
                store = get_store(aid)
                # 查找已有的该 peer 的活跃会话
                existing_sid = next(
                    (sid for sid, s in store.items()
                     if s.get('peer_aid') == inviter and not s.get('closed')),
                    None
                )
                if existing_sid and existing_sid != session_id:
                    # 把已有会话的消息迁移到对方邀请的 session 上，双方共用同一个 session
                    old = store[existing_sid]
                    store[session_id] = {
                        'session_id': session_id, 'peer_aid': inviter,
                        'type': old.get('type', 'incoming'), 'identifying_code': invite_code,
                        'messages': old.get('messages', []), 'closed': False,
                        'created_at': int(time.time() * 1000),
                        'last_message_at': old.get('last_message_at', 0),
                        'last_message': old.get('last_message', ''),
                    }
                    del store[existing_sid]
                    _rename_session_file(aid, existing_sid, session_id)
                    _flush_index(aid)
                    log_info(f'[Server] AID {aid} 会话迁移: {existing_sid[:8]}... -> {session_id[:8]}... (peer={inviter})')
                elif not existing_sid:
                    get_or_create_session(aid, session_id, inviter, 'incoming', invite_code)
                # else: session_id == existing_sid, 无需处理
                push_to_aid(aid, {'type': 'sessions_updated'})
            if original_on_invite:
                original_on_invite(invite_req)
        agent_id.heartbeat_client.set_on_recv_invite(on_invite)

    inst = AidInstance(aid, agent_id)
    inst.online = True
    inst.ws_connected = True
    inst.ws_status = 'connected'
    aid_instances[aid] = inst

    # Rebuild stale sessions: online() destroys all server-side sessions,
    # so we need to create new ones and migrate data in-place.
    store = get_store(aid)
    stale_sids = [sid for sid, s in store.items()
                  if not s.get('closed') and s.get('peer_aid')]
    if stale_sids:
        log_info(f'[Server] AID {aid} 重建 {len(stale_sids)} 个旧会话...')
        for old_sid in stale_sids:
            s = store[old_sid]
            peer = s['peer_aid']
            try:
                new_sid = agent_id.create_session(f'p2p_{int(time.time())}', 'ACP P2P')
                if not new_sid:
                    continue
                agent_id.invite_member(new_sid, peer)
                # Migrate in-place: copy data to new key, remove old
                store[new_sid] = {
                    'session_id': new_sid, 'peer_aid': peer,
                    'type': s.get('type', 'outgoing'), 'identifying_code': '',
                    'messages': s.get('messages', []), 'closed': False,
                    'created_at': int(time.time() * 1000),
                    'last_message_at': s.get('last_message_at', 0),
                    'last_message': s.get('last_message', ''),
                }
                del store[old_sid]
                _rename_session_file(aid, old_sid, new_sid)
            except Exception as e:
                log_warning(f'[Server] 重建会话 {old_sid} 失败: {e}')
        _flush_index(aid)

    log_info(f'[Server] AID {aid} 已上线')
    push_to_browser({'type': 'aid_status', 'aidStatus': get_aid_status_list()})
    return inst


def ensure_group_client(inst: AidInstance):
    """Initialize group client for an AID instance if not already done."""
    if inst.group_initialized and inst.group_client:
        return
    aid = inst.aid
    parts = aid.split('.')
    target_aid = 'group.' + ('.'.join(parts[1:]) if len(parts) >= 3 else aid.split('.', 1)[-1])
    inst.group_target_aid = target_aid

    # Create session with group server
    session_id = inst.agent_id.create_session(f'group_{int(time.time())}', 'ACP Group')
    if not session_id:
        raise Exception('Failed to create group session')
    inst.agent_id.invite_member(session_id, target_aid)
    inst.group_session_id = session_id

    # Create group client and operations
    client = GroupClient(inst.agent_id, session_id, target_aid)
    inst.group_client = client
    inst.group_ops = GroupOperations(client, target_aid)
    inst.group_initialized = True

    # Set up message push callback for real-time group messages
    def on_group_message_push(group_id, msg):
        if not group_id:
            return
        with group_message_lock:
            store = group_message_stores.setdefault(group_id, [])
            msg_id = msg.get('msg_id', 0)
            if msg_id and any(m.get('msg_id') == msg_id for m in store):
                return  # deduplicate
            store.append(msg)
            store.sort(key=lambda x: x.get('msg_id', 0))
        push_to_aid(aid, {'type': 'group_message', 'group_id': group_id, 'message': msg})

    client.set_on_message_push(on_group_message_push)

    # Try register online
    try:
        inst.group_ops.register_online()
    except Exception as e:
        log_warning(f'[Group] register_online failed: {e}')

    log_info(f'[Group] Client initialized for {aid}, target={target_aid}')


# ============================================================
# Heartbeat health check (auto-reconnect)
# ============================================================
HEALTH_CHECK_INTERVAL = 30  # seconds
_health_check_running = False


def _health_check_loop():
    """Periodically verify server-side online status and notify browser."""
    while _health_check_running:
        try:
            for aid, inst in list(aid_instances.items()):
                if not inst.online:
                    continue
                hb = inst.agent_id.heartbeat_client
                if not hb or not hb.is_running:
                    continue
                try:
                    status_list = hb.get_online_status([aid])
                    server_online = any(
                        s.get('aid') == aid and s.get('online')
                        for s in (status_list or [])
                    )
                    if not server_online:
                        log_warning(f'[HealthCheck] {aid} 服务端不在线')
                        inst.ws_status = 'disconnected'
                        push_to_browser({'type': 'ws_status', 'aid': aid, 'status': 'disconnected'})
                    elif inst.ws_status == 'disconnected':
                        inst.ws_status = 'connected'
                        push_to_browser({'type': 'ws_status', 'aid': aid, 'status': 'connected'})
                except Exception as e:
                    log_warning(f'[HealthCheck] {aid} 检查失败: {e}')
        except Exception as e:
            log_error(f'[HealthCheck] Error: {e}')
        time.sleep(HEALTH_CHECK_INTERVAL)


def start_health_check():
    global _health_check_running
    if _health_check_running:
        return
    _health_check_running = True
    t = threading.Thread(target=_health_check_loop, daemon=True)
    t.start()


# ============================================================
# Group message background polling
# ============================================================
GROUP_POLL_INTERVAL = 3  # seconds


def start_group_poll(aid: str):
    if aid in group_poll_threads and group_poll_threads[aid].is_alive():
        return
    group_poll_running[aid] = True
    t = threading.Thread(target=_group_poll_loop, args=(aid,), daemon=True)
    group_poll_threads[aid] = t
    t.start()


def stop_group_poll(aid: str):
    group_poll_running[aid] = False


def _fetch_group_list(inst: AidInstance) -> list:
    """Fetch group list from server if not cached locally."""
    aid = inst.aid
    groups = group_list_stores.get(aid, [])
    if groups or not inst.group_ops:
        return groups
    try:
        result = inst.group_ops.list_my_groups()
        for g in result.get('groups', []):
            gid = g.get('group_id', '')
            if gid:
                name = gid
                try:
                    info = inst.group_ops.get_group_info(gid)
                    name = info.get('name', gid) or gid
                except Exception:
                    pass
                groups.append({'group_id': gid, 'name': name})
        if groups:
            group_list_stores[aid] = groups
    except Exception as e:
        log_warning(f'[Group] {aid}: list_my_groups error: {e}')
    return groups


def _group_poll_loop(aid: str):
    log_info(f'[GroupPoll] Started for {aid}')
    while group_poll_running.get(aid, False):
        try:
            inst = aid_instances.get(aid)
            if not inst or not inst.group_ops:
                time.sleep(GROUP_POLL_INTERVAL)
                continue
            groups = _fetch_group_list(inst)
            if not groups:
                time.sleep(GROUP_POLL_INTERVAL)
                continue
            for g in groups:
                gid = g.get('group_id', '')
                if not gid:
                    continue
                try:
                    cursors = group_poll_cursors.setdefault(aid, {})
                    cursor = cursors.get(gid, -1)
                    if cursor == -1:
                        # 首次轮询：探测 latest_msg_id，直接跳到最新位置
                        # 只监听之后的新消息，不拉历史
                        try:
                            probe = inst.group_ops._client.send_request(
                                'pull_messages', gid, {'after_msg_id': 0, 'limit': 1}
                            )
                            latest = probe.get('data', {}).get('latest_msg_id', 0)
                            cursors[gid] = latest
                        except Exception:
                            cursors[gid] = 0
                        continue
                    resp = inst.group_ops._client.send_request(
                        'pull_messages', gid, {'after_msg_id': cursor, 'limit': 50}
                    )
                    d = resp.get('data', {})
                    new_msgs = d.get('messages', [])
                    latest = d.get('latest_msg_id', 0)
                    if latest > cursor:
                        cursors[gid] = latest
                    added = 0
                    with group_message_lock:
                        store = group_message_stores.setdefault(gid, [])
                        existing_ids = {m.get('msg_id') for m in store}
                        for m in new_msgs:
                            mid = m.get('msg_id')
                            if not mid:
                                continue
                            if mid not in existing_ids:
                                store.append(m)
                                existing_ids.add(mid)
                            if mid > cursor:
                                push_to_aid(aid, {'type': 'group_message', 'group_id': gid, 'message': m})
                                added += 1
                        if added:
                            store.sort(key=lambda x: x.get('msg_id', 0))
                except Exception as e:
                    log_warning(f'[GroupPoll] poll error {gid}: {e}')
        except Exception as e:
            log_error(f'[GroupPoll] Error for {aid}: {e}')
        time.sleep(GROUP_POLL_INTERVAL)
    log_info(f'[GroupPoll] Stopped for {aid}')


def handle_incoming_message(aid: str, data: dict):
    """Handle incoming message - routes to group client or P2P handler"""
    try:
        sender = data.get('sender', '')
        session_id = data.get('session_id', '')
        msg_content = data.get('message', '')

        # Try to decode message content
        decoded = msg_content
        if isinstance(decoded, str):
            try:
                if '%' in decoded:
                    decoded = urllib.parse.unquote(decoded)
            except Exception:
                pass

        # Try routing to group client
        inst = aid_instances.get(aid)
        if inst and inst.group_client and inst.group_session_id == session_id:
            if inst.group_client.handle_incoming(decoded):
                return

        # P2P message handling
        content = ''
        if isinstance(decoded, str):
            try:
                parsed = json.loads(decoded)
                if isinstance(parsed, list) and len(parsed) > 0:
                    content = ''.join(item.get('content', '') for item in parsed)
                elif isinstance(parsed, dict) and 'content' in parsed:
                    content = parsed['content']
                else:
                    content = decoded
            except Exception:
                content = decoded
        else:
            content = json.dumps(msg_content) if msg_content else json.dumps(data)

        if session_id and sender:
            store = get_store(aid)
            target_sid = session_id
            if session_id not in store:
                # Route to existing active session with this sender if available
                existing_sid = next(
                    (sid for sid, s in store.items()
                     if s.get('peer_aid') == sender and not s.get('closed')),
                    None
                )
                if existing_sid:
                    target_sid = existing_sid
                else:
                    get_or_create_session(aid, session_id, sender, 'incoming')
            msg = {'type': 'received', 'content': content, 'from': sender, 'timestamp': int(time.time() * 1000)}
            add_message_to_session(aid, target_sid, msg)
            push_to_aid(aid, {'type': 'p2p_message', 'sessionId': target_sid, 'message': msg})
            push_to_aid(aid, {'type': 'sessions_updated'})
    except Exception as e:
        log_error(f'[Server] handle_incoming_message error: {e}')


# ============================================================
# Flask Routes - Pages & SSE
# ============================================================
TEMPLATE_DIR = os.path.join(os.path.dirname(__file__), 'templates')
ASSETS_DIR = os.path.join(os.path.dirname(__file__), 'assets')


@app.route('/')
def index_page():
    return send_from_directory(TEMPLATE_DIR, 'index.html')


@app.route('/chat')
def chat_page():
    return send_from_directory(TEMPLATE_DIR, 'chat.html')


@app.route('/assets/<path:filename>')
def serve_asset(filename):
    allowed = ['openclaw.png', 'human.png', 'agent.png']
    if filename not in allowed:
        return 'Not Found', 404
    return send_from_directory(ASSETS_DIR, filename, max_age=86400)


def _ws_handler(websocket):
    q = queue.Queue()
    with ws_lock:
        ws_queues.append(q)
    try:
        websocket.send(json.dumps({'type': 'ws_connected'}))
        while True:
            while not q.empty():
                try:
                    msg = q.get_nowait()
                    websocket.send(msg)
                except Exception:
                    break
            try:
                websocket.recv(timeout=0.5)
            except ConnectionClosed:
                break
            except Exception:
                pass
    except ConnectionClosed:
        pass
    except Exception:
        pass
    finally:
        with ws_lock:
            if q in ws_queues:
                ws_queues.remove(q)


def _start_ws_server(port):
    with ws_serve(_ws_handler, '0.0.0.0', port) as server:
        server.serve_forever()


# ============================================================
# Flask Routes - AID Management
# ============================================================
@app.route('/api/aid')
def api_aid():
    try:
        cp = get_acp()
        aid_list = cp.get_aid_list()
        aid_status = get_aid_status_list()
        return jsonify({'aidList': aid_list, 'aidStatus': aid_status, 'apiUrl': ''})
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@app.route('/api/aid/create', methods=['POST'])
def api_aid_create():
    try:
        body = request.get_json(force=True)
        prefix = (body.get('prefix') or '').strip()
        if not prefix:
            return jsonify({'success': False, 'error': '请输入 AID 名称'})
        cp = get_acp()
        existing = cp.get_aid_list()
        if len(existing) >= MAX_AIDS:
            return jsonify({'success': False, 'error': f'最多只能注册 {MAX_AIDS} 个 AID'})
        ap = body.get('ap', 'agentcp.io')
        try:
            agent_id = cp.create_aid(ap=ap, agent_name=prefix)
            aid = agent_id.id
            nickname = (body.get('nickname') or '').strip()
            description = (body.get('description') or '').strip()
            if nickname or description:
                opts = {}
                if nickname: opts['name'] = nickname
                if description: opts['description'] = description
                save_aid_md_options_for(aid, opts)
            # Auto-online to trigger agent.md generation and upload
            try:
                ensure_online(aid)
            except Exception as e:
                log_warning(f'[Server] Auto-online after create failed for {aid}: {e}')
            return jsonify({'success': True, 'aid': aid})
        except Exception as e:
            return jsonify({'success': False, 'error': f'AID "{prefix}" 注册失败，该名称可能已被占用'})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/aid/select', methods=['POST'])
def api_aid_select():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        if not aid:
            return jsonify({'success': False, 'error': '请指定 AID'})
        try:
            ensure_online(aid)
            return jsonify({'success': True, 'aid': aid})
        except Exception as e:
            return jsonify({'success': False, 'error': str(e)})
    except Exception as e:
        return jsonify({'error': str(e)}), 500


@app.route('/api/aid/validate', methods=['POST'])
def api_aid_validate():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        if not aid:
            return jsonify({'success': False, 'error': '请指定 AID'})
        status = validate_aid(aid)
        inst = aid_instances.get(aid)
        return jsonify({'success': True, 'aid': aid, **status, 'online': inst.online if inst else False})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/aid/offline', methods=['POST'])
def api_aid_offline():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        if not aid:
            return jsonify({'success': False, 'error': '请指定 AID'})
        inst = aid_instances.get(aid)
        if not inst:
            return jsonify({'success': False, 'error': '该 AID 未上线'})
        try:
            inst.agent_id.offline()
        except Exception:
            pass
        stop_group_poll(aid)
        del aid_instances[aid]
        log_info(f'[Server] AID {aid} 已下线')
        push_to_browser({'type': 'aid_status', 'aidStatus': get_aid_status_list()})
        return jsonify({'success': True, 'aid': aid})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


# ============================================================
# Flask Routes - Agent Info
# ============================================================
@app.route('/api/agent-info')
def api_agent_info():
    aid = request.args.get('aid', '')
    if not aid:
        return jsonify({'type': '', 'name': '', 'description': '', 'tags': []})
    return jsonify(get_agent_info(aid))


@app.route('/api/agent-md-raw')
def api_agent_md_raw():
    aid = request.args.get('aid', '')
    if not aid:
        return jsonify({'success': False, 'error': '缺少 aid'})
    try:
        md = fetch_agent_md(aid)
        return jsonify({'success': True, 'content': md})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


# ============================================================
# Flask Routes - WebSocket / P2P
# ============================================================
@app.route('/api/ws/start', methods=['POST'])
def api_ws_start():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        if not aid:
            return jsonify({'success': False, 'error': '请先选择 AID'})
        ensure_online(aid)
        return jsonify({'success': True, 'aid': aid})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/ws/connect', methods=['POST'])
def api_ws_connect():
    try:
        body = request.get_json(force=True)
        target_aid = body.get('targetAid', '')
        aid = body.get('aid', '')
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        if not target_aid:
            return jsonify({'success': False, 'error': '缺少目标 AID'})
        inst = ensure_online(aid)
        agent_id = inst.agent_id
        session_id = agent_id.create_session(f'p2p_{int(time.time())}', 'ACP P2P')
        if not session_id:
            return jsonify({'success': False, 'error': '创建会话失败'})
        agent_id.invite_member(session_id, target_aid)
        get_or_create_session(aid, session_id, target_aid, 'outgoing')
        push_to_aid(aid, {'type': 'sessions_updated'})
        return jsonify({'success': True, 'sessionId': session_id, 'identifyingCode': ''})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/ws/send', methods=['POST'])
def api_ws_send():
    try:
        body = request.get_json(force=True)
        message = body.get('message', '')
        session_id = body.get('sessionId', '')
        aid = body.get('aid', '')
        if not message:
            return jsonify({'success': False, 'error': '消息不能为空'})
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        if not session_id:
            return jsonify({'success': False, 'error': '缺少 sessionId'})
        inst = ensure_online(aid)
        store = get_store(aid)
        session = store.get(session_id)
        if not session:
            return jsonify({'success': False, 'error': '会话不存在'})
        if session.get('closed'):
            return jsonify({'success': False, 'error': '该会话已关闭，请新建会话'})
        peer_aid = session['peer_aid']
        inst.agent_id.send_message_content(session_id, [peer_aid], message)
        msg = {'type': 'sent', 'content': message, 'to': peer_aid, 'timestamp': int(time.time() * 1000)}
        add_message_to_session(aid, session_id, msg)
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/ws/status')
def api_ws_status():
    aid = request.args.get('aid', '')
    inst = aid_instances.get(aid) if aid else None
    if not inst:
        return jsonify({'connected': False, 'status': 'disconnected'})
    return jsonify({'connected': inst.ws_connected, 'status': inst.ws_status})


# ============================================================
# Flask Routes - Sessions
# ============================================================
@app.route('/api/sessions')
def api_sessions():
    aid = request.args.get('aid', '')
    if not aid:
        return jsonify({'sessions': [], 'activeSessionId': None})
    return jsonify({'sessions': get_session_list(aid), 'activeSessionId': None})


@app.route('/api/messages')
def api_messages():
    aid = request.args.get('aid', '')
    session_id = request.args.get('sessionId', '')
    if not aid or not session_id:
        return jsonify({'messages': [], 'activeSessionId': None, 'closed': False})
    store = get_store(aid)
    session = store.get(session_id)
    return jsonify({
        'messages': session['messages'] if session else [],
        'activeSessionId': session_id,
        'closed': session.get('closed', False) if session else False
    })


@app.route('/api/sessions/active', methods=['POST'])
def api_sessions_active():
    try:
        body = request.get_json(force=True)
        session_id = body.get('sessionId', '')
        aid = body.get('aid', '')
        if not aid or not session_id:
            return jsonify({'success': False, 'error': '会话不存在'})
        store = get_store(aid)
        if session_id not in store:
            return jsonify({'success': False, 'error': '会话不存在'})
        push_to_aid(aid, {'type': 'set_active_session', 'sessionId': session_id})
        return jsonify({'success': True, 'activeSessionId': session_id})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/sessions/delete', methods=['POST'])
def api_sessions_delete():
    try:
        body = request.get_json(force=True)
        session_id = body.get('sessionId', '')
        aid = body.get('aid', '')
        if not session_id:
            return jsonify({'success': False, 'error': '会话ID不能为空'})
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        store = get_store(aid)
        if session_id in store:
            del store[session_id]
            # Clean up files
            try:
                msg_path = os.path.join(_sessions_dir(aid), f'{session_id}.jsonl')
                if os.path.exists(msg_path):
                    os.remove(msg_path)
            except Exception:
                pass
            _flush_index(aid)
            return jsonify({'success': True})
        return jsonify({'success': False, 'error': '会话不存在或删除失败'})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/peers/delete', methods=['POST'])
def api_peers_delete():
    try:
        body = request.get_json(force=True)
        peer_aid = body.get('peerAid', '')
        aid = body.get('aid', '')
        if not peer_aid:
            return jsonify({'success': False, 'error': 'AID不能为空'})
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        store = get_store(aid)
        count = 0
        to_del = [sid for sid, s in store.items() if s['peer_aid'] == peer_aid]
        for sid in to_del:
            del store[sid]
            try:
                msg_path = os.path.join(_sessions_dir(aid), f'{sid}.jsonl')
                if os.path.exists(msg_path):
                    os.remove(msg_path)
            except Exception:
                pass
            count += 1
        if to_del:
            _flush_index(aid)
        return jsonify({'success': True, 'count': count})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


# ============================================================
# Flask Routes - Group
# ============================================================
@app.route('/api/group/init', methods=['POST'])
def api_group_init():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        start_group_poll(aid)
        groups = _fetch_group_list(inst)
        active = inst.active_group_id if inst else None
        return jsonify({'success': True, 'targetAid': inst.group_target_aid, 'groups': groups, 'activeGroupId': active})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/create', methods=['POST'])
def api_group_create():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        name = body.get('name', '')
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        if not name:
            return jsonify({'success': False, 'error': '群组名称不能为空'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.create_group(
            name,
            visibility=body.get('visibility', ''),
            description=body.get('description', ''),
        )
        group_id = result.get('group_id', '')
        # Set duty mode if provided
        duty_mode = body.get('duty_mode', '')
        if duty_mode and group_id:
            try:
                inst.group_ops.update_duty_config(group_id, {'mode': duty_mode})
            except Exception:
                pass
        # Store locally
        if group_id:
            groups = group_list_stores.setdefault(aid, [])
            groups.append({'group_id': group_id, 'name': name})
        return jsonify({'success': True, **result})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/list')
def api_group_list():
    try:
        aid = request.args.get('aid', '')
        if not aid:
            return jsonify({'success': True, 'groups': [], 'activeGroupId': None})
        inst = aid_instances.get(aid)
        groups = _fetch_group_list(inst) if inst else []
        active = inst.active_group_id if inst else None
        return jsonify({'success': True, 'groups': groups, 'activeGroupId': active})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e), 'groups': []})


@app.route('/api/group/select', methods=['POST'])
def api_group_select():
    try:
        body = request.get_json(force=True)
        aid = body.get('aid', '')
        group_id = body.get('groupId', '')
        inst = aid_instances.get(aid)
        if inst:
            inst.active_group_id = group_id
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/info')
def api_group_info():
    try:
        group_id = request.args.get('groupId', '')
        aid = request.args.get('aid', '')
        if not group_id or not aid:
            return jsonify({'success': False, 'error': '缺少参数'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        info = inst.group_ops.get_group_info(group_id)
        return jsonify({'success': True, **info})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/send', methods=['POST'])
def api_group_send():
    try:
        body = request.get_json(force=True)
        group_id = body.get('groupId', '')
        message = body.get('message', '')
        aid = body.get('aid', '')
        if not group_id or not message:
            return jsonify({'success': False, 'error': '缺少 groupId 或 message'})
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.send_message(group_id, message)
        # Store locally
        with group_message_lock:
            msgs = group_message_stores.setdefault(group_id, [])
            msgs.append({
                'msg_id': result.get('msg_id', 0),
                'sender': aid,
                'content': message,
                'content_type': 'text',
                'timestamp': result.get('timestamp', int(time.time() * 1000)),
            })
        return jsonify({'success': True, **result})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/messages')
def api_group_messages():
    try:
        group_id = request.args.get('groupId', '')
        aid = request.args.get('aid', '')
        before_msg_id = int(request.args.get('before_msg_id', '0') or '0')
        limit = int(request.args.get('limit', '0') or '0')
        fetch_limit = limit if limit > 0 else 30
        if not aid or not group_id:
            return jsonify({'success': True, 'messages': [], 'has_more': False})
        inst = aid_instances.get(aid)
        with group_message_lock:
            store = group_message_stores.setdefault(group_id, [])

            if inst and inst.group_ops:
                if before_msg_id > 0:
                    # 向上翻页：拉 before_msg_id 之前的消息
                    local_oldest = min((m.get('msg_id', 0) for m in store), default=0) if store else 0
                    if not store or local_oldest >= before_msg_id:
                        try:
                            fetch_after = max(0, before_msg_id - fetch_limit - 1)
                            result = inst.group_ops.pull_messages(group_id, after_msg_id=fetch_after, limit=fetch_limit)
                            _merge_to_store(store, result.get('messages', []))
                        except Exception:
                            pass
                else:
                    # 初次加载：始终从服务端拉最新一页
                    try:
                        probe = inst.group_ops.pull_messages(group_id, after_msg_id=0, limit=1)
                        latest = probe.get('latest_msg_id', 0)
                        if latest > 0:
                            start = max(0, latest - fetch_limit)
                            result = inst.group_ops.pull_messages(group_id, after_msg_id=start, limit=fetch_limit)
                            _merge_to_store(store, result.get('messages', []))
                        else:
                            _merge_to_store(store, probe.get('messages', []))
                    except Exception:
                        pass

            # 分页返回
            msgs = list(store)
        if before_msg_id > 0:
            msgs = [m for m in msgs if m.get('msg_id', 0) < before_msg_id]
        has_more = False
        if len(msgs) > fetch_limit:
            has_more = True
            msgs = msgs[-fetch_limit:]
        elif msgs:
            # 返回的最早一条 msg_id > 1 说明服务端还有更早的消息
            has_more = msgs[0].get('msg_id', 0) > 1
        return jsonify({'success': True, 'messages': msgs, 'has_more': has_more})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e), 'messages': [], 'has_more': False})


def _merge_to_store(store: list, new_msgs: list):
    if not new_msgs:
        return
    existing_ids = {m.get('msg_id') for m in store}
    for m in new_msgs:
        if m.get('msg_id') and m['msg_id'] not in existing_ids:
            store.append(m)
            existing_ids.add(m['msg_id'])
    store.sort(key=lambda x: x.get('msg_id', 0))


@app.route('/api/group/invite-code', methods=['POST'])
def api_group_invite_code():
    try:
        body = request.get_json(force=True)
        group_id = body.get('groupId', '')
        aid = body.get('aid', '')
        if not group_id or not aid:
            return jsonify({'success': False, 'error': '缺少参数'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.create_invite_code(group_id)
        group_url = f'https://{inst.group_target_aid}/{group_id}'
        return jsonify({'success': True, **result, 'group_url': group_url})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


def _finalize_group_join(inst: AidInstance, group_id: str):
    """After joining a group, fetch name and store locally."""
    name = group_id
    try:
        info = inst.group_ops.get_group_info(group_id)
        name = info.get('name', group_id)
    except Exception:
        pass
    groups = group_list_stores.setdefault(inst.aid, [])
    if not any(g['group_id'] == group_id for g in groups):
        groups.append({'group_id': group_id, 'name': name})


@app.route('/api/group/join', methods=['POST'])
def api_group_join():
    try:
        body = request.get_json(force=True)
        group_url = body.get('groupUrl', '')
        code = body.get('code', '')
        aid = body.get('aid', '')
        if not group_url:
            return jsonify({'success': False, 'error': '缺少群聊链接'})
        if not aid:
            return jsonify({'success': False, 'error': '缺少 aid'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        target_aid, group_id = GroupOperations.parse_group_url(group_url)
        if code:
            inst.group_ops.use_invite_code(group_id, code)
            _finalize_group_join(inst, group_id)
            return jsonify({'success': True, 'group_id': group_id})
        else:
            result = inst.group_ops.request_join(group_id, body.get('message', ''))
            if result['status'] == 'joined':
                _finalize_group_join(inst, group_id)
                return jsonify({'success': True, 'group_id': group_id})
            else:
                return jsonify({'success': True, 'pending': True, 'request_id': result['request_id']})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/pending-requests')
def api_group_pending_requests():
    try:
        group_id = request.args.get('groupId', '')
        aid = request.args.get('aid', '')
        if not group_id or not aid:
            return jsonify({'success': True, 'requests': []})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.get_pending_requests(group_id)
        return jsonify({'success': True, **result})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e), 'requests': []})


@app.route('/api/group/review-join', methods=['POST'])
def api_group_review_join():
    try:
        body = request.get_json(force=True)
        group_id = body.get('groupId', '')
        agent_id = body.get('agentId', '')
        action = body.get('action', '')
        aid = body.get('aid', '')
        if not group_id or not agent_id or not action or not aid:
            return jsonify({'success': False, 'error': '缺少参数'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        inst.group_ops.review_join_request(group_id, agent_id, action, body.get('reason', ''))
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/members')
def api_group_members():
    try:
        group_id = request.args.get('groupId', '')
        aid = request.args.get('aid', '')
        if not group_id or not aid:
            return jsonify({'success': True, 'members': []})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.get_members(group_id)
        return jsonify({'success': True, **result})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e), 'members': []})


@app.route('/api/group/my-groups')
def api_group_my_groups():
    try:
        aid = request.args.get('aid', '')
        if not aid:
            return jsonify({'success': True, 'groups': [], 'total': 0})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.list_my_groups()
        groups = []
        for m in result.get('groups', []):
            name = m.get('group_id', '')
            try:
                info = inst.group_ops.get_group_info(m['group_id'])
                name = info.get('name', name)
            except Exception:
                pass
            groups.append({**m, 'name': name})
        return jsonify({'success': True, 'groups': groups, 'total': result.get('total', 0)})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e), 'groups': []})


@app.route('/api/group/leave', methods=['POST'])
def api_group_leave():
    try:
        body = request.get_json(force=True)
        group_id = body.get('groupId', '')
        aid = body.get('aid', '')
        if not group_id or not aid:
            return jsonify({'success': False, 'error': '缺少参数'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        inst.group_ops.leave_group(group_id)
        # Remove from local store
        groups = group_list_stores.get(aid, [])
        group_list_stores[aid] = [g for g in groups if g['group_id'] != group_id]
        group_message_stores.pop(group_id, None)
        if inst.active_group_id == group_id:
            inst.active_group_id = None
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/duty-status')
def api_group_duty_status():
    try:
        aid = request.args.get('aid', '')
        group_id = request.args.get('groupId', '')
        if not aid or not group_id:
            return jsonify({'success': True, 'config': {}, 'state': {}})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        result = inst.group_ops.get_duty_status(group_id)
        return jsonify({'success': True, **result})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


@app.route('/api/group/update-duty-config', methods=['POST'])
def api_group_update_duty_config():
    try:
        body = request.get_json(force=True)
        group_id = body.get('groupId', '')
        aid = body.get('aid', '')
        mode = body.get('mode', '')
        if not aid or not group_id or not mode:
            return jsonify({'success': False, 'error': '缺少参数'})
        inst = ensure_online(aid)
        ensure_group_client(inst)
        inst.group_ops.update_duty_config(group_id, {'mode': mode})
        return jsonify({'success': True})
    except Exception as e:
        return jsonify({'success': False, 'error': str(e)})


# ============================================================
# Server Startup
# ============================================================
def start_server(port: int = 9527, data_dir: str = '', seed_password: str = '', open_browser: bool = True):
    global global_data_dir, global_seed_password
    global_data_dir = os.path.abspath(data_dir) if data_dir else os.path.join(os.getcwd(), 'acp-data')
    # 密码校验：参数 > 环境变量，都没有则拒绝启动
    if not seed_password:
        seed_password = os.environ.get('ACP_SEED_PASSWORD', '')
    if not seed_password:
        raise ValueError('seed_password 未设置，请通过参数传入或设置环境变量 ACP_SEED_PASSWORD')
    global_seed_password = seed_password
    os.makedirs(global_data_dir, exist_ok=True)
    load_aid_md_options()
    ws_port = port + 1
    print(f'ACP Web Server starting on http://localhost:{port}')
    print(f'WebSocket server starting on ws://localhost:{ws_port}')
    print(f'Data dir: {global_data_dir}')
    threading.Thread(target=_start_ws_server, args=(ws_port,), daemon=True).start()
    start_health_check()
    if open_browser:
        def _open_browser():
            import webbrowser
            time.sleep(1)
            webbrowser.open(f'http://localhost:{port}')
        threading.Thread(target=_open_browser, daemon=True).start()
    try:
        app.run(host='0.0.0.0', port=port, threaded=True)
    except OSError as e:
        if 'Address already in use' in str(e) or getattr(e, 'errno', 0) == 48:
            print(f'\n端口 {port} 已被占用，请使用 --port 指定其他端口，或先关闭占用该端口的进程')
        else:
            print(f'\n启动失败: {e}')
    except KeyboardInterrupt:
        print('\n正在退出...')



# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import os
import shutil
import time

import requests
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509.oid import NameOID

from agentcp.base.env import Environ
from agentcp.base.log import log_debug, log_error, log_exception, log_info
from agentcp.utils import tls_config as tls

from ..context import ErrorContext, exceptions


class CAClient:
    def __init__(self, ca_server, aid_path, seed_password: str, timeout=30):
        self.ca_server = ca_server or Environ.CA_SERVER.get()  # 移除末尾的斜杠
        if not self.ca_server or not self.ca_server.startswith(("http://", "https://")):
            raise ValueError("无效的CA服务器地址")

        self.ca_server = self.ca_server.rstrip("/")  # 移除末尾的斜杠
        self.ca_server = self.ca_server + "/api/accesspoint"
        self.timeout = timeout
        self.seed_password = seed_password
        self.aid_path = aid_path

    def get_aid_certs_path(self, aid_str):
        return os.path.join(self.aid_path, aid_str, "private", "certs")

    def __save_csr_to_file(self, csr, filename):
        try:
            # 确保目录存在
            file_dir = os.path.dirname(filename)
            os.makedirs(file_dir, exist_ok=True)
            with open(filename, "wb") as f:
                f.write(csr.public_bytes(serialization.Encoding.PEM))
            log_debug(f"CSR save to {filename}")  # 调试用
        except Exception as e:
            log_exception(f"save csr to file error: {e}")  # 调试用

    def save_cert_to_file(self, name, certificate: str):
        try:
            aid_path = os.path.join(self.aid_path, name, "private", "certs")
            # 确保目录存在
            os.makedirs(aid_path, exist_ok=True)
            aid_path = os.path.join(aid_path, name + ".crt")
            with open(aid_path, "wb") as f:
                f.write(certificate.encode("utf-8"))
            log_debug(f"private key saveto {name}.key")  # 调试用
        except Exception as e:
            log_exception(f"save private key to file error{e}")  # 调试用
            ErrorContext.publish(exceptions.SDKError(f"保存证书失败: {e}"))
            raise RuntimeError("保存证书失败")

    def save_private_key_to_file(self, name, private_key):
        try:
            aid_path = os.path.join(self.aid_path, name, "private", "certs")
            # 确保目录存在
            os.makedirs(aid_path, exist_ok=True)
            aid_path = os.path.join(aid_path, name + ".key")
            with open(aid_path, "wb") as f:
                f.write(
                    private_key.private_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PrivateFormat.PKCS8,
                        encryption_algorithm=serialization.BestAvailableEncryption(self.seed_password.encode()),
                    )
                )
            log_debug(f"private key saveto {name}.key")  # 调试用
        except Exception as e:
            log_exception(f"save private key to file error{e}")  # 调试用
            ErrorContext.publish(exceptions.SDKError(f"保存私钥失败: {e}"))
            raise RuntimeError("保存私钥失败")

    def modify_seed_password(self, aid_str, private_key, new_send_password):
        temp_path = os.path.join(self.aid_path, aid_str)
        os.path.exists(temp_path) or os.makedirs(temp_path)
        aid_path = os.path.join(temp_path, aid_str + ".key")
        with open(aid_path, "wb") as f:
            f.write(
                private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.BestAvailableEncryption(new_send_password.encode("utf-8")),
                )
            )
        old_aid_path = os.path.join(self.aid_path, aid_str, "private", aid_str + ".key")
        old_key_path = os.path.join(self.aid_path, aid_str, "private", "old.key")
        os.rename(old_aid_path, old_key_path)
        # 复制新文件到旧路径
        shutil.copy2(aid_path, old_aid_path)
        # 删除old.key
        os.remove(old_key_path)
        os.remove(aid_path)

    def __generate_private_key(self):
        """
        生成NIST P-384椭圆曲线私钥
        :return: 返回生成的私钥对象
        """
        # 使用SECP384R1曲线生成私钥
        private_key = ec.generate_private_key(ec.SECP384R1())
        return private_key

    def __generate_csr(self, private_key, common_name):
        """
        使用NIST P-384私钥生成证书签名请求(CSR)
        :param private_key: NIST P-384椭圆曲线私钥
        :param common_name: 证书通用名称
        :return: 返回生成的CSR对象
        """
        # 创建 CSR 的主体信息
        csr_builder = x509.CertificateSigningRequestBuilder().subject_name(
            x509.Name(
                [
                    x509.NameAttribute(NameOID.COUNTRY_NAME, "CN"),
                    x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, "SomeState"),
                    x509.NameAttribute(NameOID.LOCALITY_NAME, "SomeCity"),
                    x509.NameAttribute(NameOID.ORGANIZATION_NAME, "SomeOrganization"),
                    x509.NameAttribute(NameOID.COMMON_NAME, common_name),
                ]
            )
        )

        # 添加扩展（可选）
        csr_builder = csr_builder.add_extension(
            x509.BasicConstraints(ca=False, path_length=None),
            critical=True,
        )

        # 使用私钥对 CSR 进行签名
        csr = csr_builder.sign(private_key, hashes.SHA256(), default_backend())
        return csr

    def __load_csr(self, agent_id):
        aid_path = os.path.join(self.aid_path, agent_id, "private", "certs", agent_id + ".csr")
        if os.path.exists(aid_path):
            with open(aid_path, "rb") as f:
                csr = x509.load_pem_x509_csr(f.read())
            return csr
        return None

    def load_private_key_str(self, agent_id, seed_password: str):
        try:
            aid_path = os.path.join(self.aid_path, agent_id, "private", "certs", agent_id + ".key")
            with open(aid_path, "rb") as f:
                private_key = serialization.load_pem_private_key(
                    f.read(),
                    password=seed_password.encode("utf-8"),
                )
                return private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.BestAvailableEncryption(seed_password.encode("utf-8")),
                ).decode("utf-8")
        except Exception as e:
            ErrorContext.publish(exceptions.SDKError(f"load_private_key_str: {e}"))
            return ""

    # def load_cert_str(self,agent_id):
    #     try:
    #         aid_path = os.path.join(self.aid_path,agent_id,'private','certs',agent_id+".crt")
    #         with open(aid_path, "rb") as f:
    #             certificate_pem = f.read().decode('utf-8')
    #         return certificate_pem
    #     except Exception as e:
    #         return ""

    def load_private_key(self, agent_id):
        try:
            # 加载私钥
            aid_path = os.path.join(self.aid_path, agent_id, "private", "certs", agent_id + ".key")
            with open(aid_path, "rb") as f:
                private_key = serialization.load_pem_private_key(
                    f.read(),
                    password=self.seed_password.encode("utf-8"),
                )
            return private_key
        except Exception as e:
            # 兼容性代码，按照不加密获取private_key
            ErrorContext.publish(exceptions.SDKError(f"load_private_key: {e}"))
            return None

    def __load_public_key_pem(self, public_key):
        public_key_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode("utf-8")
        return public_key_pem

    def load_certificate_pem(self, agent_id):
        aid_path = os.path.join(self.aid_path, agent_id, "private", "certs", agent_id + ".crt")
        with open(aid_path, "rb") as f:
            certificate_pem = f.read().decode("utf-8")
        return certificate_pem

    def __get_guest_aid(self):
        os.path.exists(self.aid_path) or os.makedirs(self.aid_path)
        for entry in os.scandir(self.aid_path):
            array = entry.name.split(".")
            if entry.is_dir() and entry.name.startswith("guest"):
                return entry.name
        return ""

    def get_guest_aid(self):
        try:
            local_guest_aid = self.__get_guest_aid()
            path = os.path.join(self.aid_path, local_guest_aid, "private", "certs", local_guest_aid + ".crt")
            if local_guest_aid and self.__pen_is_valid(path):
                return local_guest_aid
            elif local_guest_aid:
                # 删除这个目录下的所有文件和子目录
                shutil.rmtree(os.path.join(self.aid_path, local_guest_aid))
            url = self.ca_server + "/sign_guest_cert"
            response = requests.get(url, verify=tls.requests_verify(), proxies={})
            if response.status_code == 200:
                rjs = response.json()
                if "guest_aid" in rjs and "key" in rjs and "cert" in rjs and "entrypoint" in rjs:
                    log_info(f"sign_guest_cert ok: guest_aid={rjs.get('guest_aid')}, entrypoint={rjs.get('entrypoint')}")
                    guest_aid = rjs["guest_aid"]
                    guest_key = rjs["key"]
                    guest_cert = rjs["cert"]
                    # 打印 guest_cert 内容，检查是否包含正确的 CSR 标记

                    try:
                        # 尝试加载证书
                        certificate = x509.load_pem_x509_certificate(guest_cert.encode("utf-8"), default_backend())
                    except Exception as e:
                        log_error(f"加载证书失败: {e}")
                        return None

                    guest_key = serialization.load_pem_private_key(guest_key.encode("utf-8"), password=None)

                    if not os.path.exists(os.path.join(self.aid_path, guest_aid)):
                        os.makedirs(os.path.join(self.aid_path, guest_aid), exist_ok=True)

                    self.save_private_key_to_file(guest_aid, guest_key)

                    # 保存证书文件
                    aid_path = os.path.join(self.aid_path, guest_aid, "private", "certs")
                    cert_name = os.path.join(aid_path, guest_aid + ".crt")
                    with open(cert_name, "wb") as f:
                        f.write(guest_cert.encode("utf-8"))
                    time.sleep(0.5)
                    return guest_aid
                else:
                    log_error(f"sign_guest_cert failed:{rjs}")
                    return None
            else:
                log_error(f"sign_guest_cert failed:{response.status_code} - {response.json().get('error', '')}")
                return None
        except Exception as e:
            ErrorContext.publish(exceptions.SDKError(f"获取访问身份失败: {e}"))
            log_error(f"获取访问身份失败: {e}")
            # log_error("详细堆栈信息:")
            # traceback.print_exc()
            return None

    def __pen_is_valid(self, agent_id_key_path: str):
        cert_valid = False
        try:
            aid_path = agent_id_key_path
            certificate_pem = ""
            with open(aid_path, "rb") as f:
                certificate_pem = f.read().decode("utf-8")
            certificate = x509.load_pem_x509_certificate(certificate_pem.encode("utf-8"), default_backend())
            # 获取证书的有效期
            not_valid_before = certificate.not_valid_before_utc
            not_valid_after = certificate.not_valid_after_utc
            # 获取当前时间
            current_time = datetime.datetime.now(datetime.timezone.utc)
            # 检查证书是否过期
            if current_time < not_valid_before:
                log_error("证书尚未生效")
            elif current_time > not_valid_after:
                log_error("证书已过期")
            elif current_time + (not_valid_after - not_valid_before) / 10 > not_valid_after:
                log_error("证书剩余有效期不足一半，需要续签")
            else:
                log_info("证书在有效期内")
                cert_valid = True
        except Exception as e:
            log_error(f"解析证书时出错: {e}")
            ErrorContext.publish(exceptions.SDKError(f"解析证书时出错: {e}"))
            cert_valid = False
        return cert_valid

    def resign_csr(self, agent_id) -> bool:
        path = os.path.join(self.aid_path, agent_id, "private", "certs", agent_id + ".crt")
        if self.__pen_is_valid(path):
            return True
        # 从CSR中提取公钥
        csr = self.__generate_csr(self.load_private_key(agent_id), agent_id)
        if csr == None:
            raise Exception("读取csr证书文件失败")
            
        public_key = csr.public_key()
        private_key = self.load_private_key(agent_id)
        public_key_pem = self.__load_public_key_pem(public_key)

        # 加载原有的证书文件
        certificate_pem = self.load_certificate_pem(agent_id)

        # 获取当前Unix时间戳(毫秒)
        current_time_ms = int(datetime.datetime.now().timestamp() * 1000)

        # 准备发送给服务器的数据
        data = {"id": agent_id, "request_id": f"{current_time_ms}", "public_key": public_key_pem}

        # 发送到服务器
        response = requests.post(f"{self.ca_server}/resign_cert", json=data, verify=tls.requests_verify(), proxies={})
        log_info(f"resign cert response: {response.content}")
        if response.status_code == 200:
            if "nonce" in response.json():
                nonce = response.json()["nonce"]
                if nonce:
                    # 使用私钥对[公钥+盐]签名，以使服务器信任私钥仍然有效
                    # 使用NIST P-384私钥对nonce进行签名
                    signature = private_key.sign((public_key_pem + nonce).encode("utf-8"), ec.ECDSA(hashes.SHA256()))
                    csr_pem = csr.public_bytes(serialization.Encoding.PEM).decode("utf-8")
                    data = {
                        "id": agent_id,
                        "request_id": f"{current_time_ms}",
                        "public_key": public_key_pem,
                        "nonce": nonce,
                        "csr": csr_pem,
                        "cert": certificate_pem,
                        "signature": signature.hex(),  # 将签名转为十六进制字符串
                    }

                    # 发送到服务器

                    response = requests.post(self.ca_server + "/resign_cert", json=data, verify=tls.requests_verify(), proxies={})
                    log_debug(f"resign cert response: {response.content}")
                    if response.status_code == 200:
                        entrypoint = ";"
                        if "entrypoint" in response.json():
                            entrypoint = response.json()["entrypoint"]
                        aid_path = os.path.join(self.aid_path, agent_id, "private", "certs", agent_id + ".crt")
                        with open(aid_path, "wb") as f:
                            f.write(response.json()["certificate"].encode("utf-8"))  # 从JSON响应中获取证书内容
                        return True
                    else:
                        log_error(f"resign csr failed: {response.status_code} - {response.json()['error']}")  # 调试用
                        return False
            else:
                log_info(
                    f"verify public key failed: {response.status_code} - {response.json().get('error', '')}"
                )  # 调试用
                ErrorContext.publish(
                    exceptions.SDKError(
                        f"verify public key failed: {response.status_code} - {response.json().get('error', '')}"
                    )
                )
                return False

    def send_csr_to_server(self, agent_id: str) -> bool:
        # 确保证书文件路径存在
        try:
            # 确保目录存在
            aid_path = os.path.join(self.aid_path, agent_id, "private", "certs")
            os.makedirs(aid_path, exist_ok=True)  # 确保aid/name目录存在

            private_key = self.__generate_private_key()
            csr = self.__generate_csr(private_key, agent_id)

            csr_pem = csr.public_bytes(serialization.Encoding.PEM).decode("utf-8")
            data = {"id": agent_id, "csr": csr_pem}
            response = requests.post(f"{self.ca_server}/sign_cert", json=data, verify=tls.requests_verify(), proxies={})
            if response.status_code == 200:
                # 确保目录存在后再保存文件
                if not os.path.exists(aid_path):
                    os.makedirs(aid_path, exist_ok=True)
                crt_name = os.path.join(aid_path, agent_id + ".crt")
                with open(crt_name, "wb") as f:
                    f.write(response.json()["certificate"].encode("utf-8"))
                    csr_name = os.path.join(aid_path, agent_id + ".csr")
                    self.__save_csr_to_file(csr, csr_name)
                    self.save_private_key_to_file(agent_id, private_key)
                log_info(f"signed certificate successfully: {agent_id} {csr_name}")  # 调试用
                return True
            else:
                delete_path = os.path.join(self.aid_path, agent_id)
                shutil.rmtree(delete_path)
                log_error(f"sign failed: {self.ca_server}, {response.status_code} - {response.text}")  # 调试用
                return response.json()["error"]
        except requests.RequestException as e:
            ErrorContext.publish(exceptions.SDKError(f"send csr to server error: {e}"))
            log_exception("send csr to server error")  # 调试用
            raise RuntimeError("send csr to server error")

    def aid_is_not_exist(self, agent_id):
        path = self.aid_path
        for entry in os.scandir(path):
            if entry.is_dir() and entry.name == agent_id:
                return False
        return True



# -*- coding: utf-8 -*-
"""
全局监控服务单例

提供跨模块访问监控服务的能力，支持AgentID进程和Server进程共享监控数据
"""

_global_monitoring_service = None


def set_global_monitoring_service(monitoring_service):
    """设置全局监控服务实例

    Args:
        monitoring_service: MonitoringService 实例
    """
    global _global_monitoring_service
    _global_monitoring_service = monitoring_service


def get_global_monitoring_service():
    """获取全局监控服务实例

    Returns:
        MonitoringService 实例，如果未设置则返回 None
    """
    return _global_monitoring_service



# 在Python文件开头明确指定编码声明
# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# from flask import Flask, request, Response, jsonify
# import threading
# import socket
# import time

# from .llm_agent_utils import LLMAgent,AttrDict
# from flask import jsonify, make_response
# import logging
# log = logging.getLogger('werkzeug')
# log.setLevel(logging.ERROR)
# app = Flask(__name__)
# # app.logger.disabled = True
# actual_port = 0

# llm_aid_app_key_map = {}
# llm_app_key_aid_map = {}
# is_running = False

# @app.route('/<llm_aid>/chat/completions', methods=['POST'])  # 添加methods参数指定POST方法
# async def llm_request(llm_aid):
#     # 获取请求头并打印
#     headers = dict(request.headers)
#     if request.is_json:
#         body = request.get_json()
#     else:
#         body = request.form.to_dict()
#     global llm_app_key_aid_map
#     auth_str:str = headers.get("Authorization")
#     llm_app_key = auth_str.replace("Bearer ","")
#     aid = llm_app_key_aid_map.get(llm_app_key)
#     if aid is None:
#         return make_response(jsonify({"error": "Unauthorized"}), 401)
#     llm_agent = LLMAgent(llm_agent=llm_aid, aid = aid)
#     response = await llm_agent.chat_create(body)
#     # print(response.get("status",""))
#     if isinstance(response, AttrDict) and response.get("status","") == 'error':
#         # 如果是错误状态，可以进行特殊处理，例如记录日志或返回自定义错误信息
#         return make_response(jsonify({"error": response.get('message', "未知错误")}), response.get('code', 400))
#     return response

# @app.route('/', defaults={'path': ''}, methods=['OPTIONS', 'GET', 'POST', 'PUT', 'DELETE', 'PATCH'])
# @app.route('/<path:path>', methods=['OPTIONS', 'GET', 'POST', 'PUT', 'DELETE', 'PATCH'])
# async def proxy(path):
#     try:
#         if request.method == 'OPTIONS':
#             # 返回 CORS 预检响应头
#             response = Response()
#             response.headers['Access-Control-Allow-Origin'] = '*'
#             response.headers['Access-Control-Allow-Methods'] = 'POST, OPTIONS'
#             response.headers['Access-Control-Allow-Headers'] = 'Content-Type, Authorization'
#             return response
#         elif request.method == 'POST':
#             # 这个最终来自路由服务器,即模型名称对应一个地址和key
#             # 校验请求数据
#             # data = request.get_json()
#             try:
#                 #return llm_request(llm_aid)
#                 llm_aid = path.split('/')[0]
#                 return await llm_request(llm_aid)
#             except Exception as e:
#                 return jsonify({"error": f"{path}"}), 502
#         elif request.method == 'GET':
#             return jsonify({"result": "服务访问正常"}), 200
#         else:
#             return jsonify({"error": "Method not allowed"}), 405

#     except Exception as e:
#         return jsonify({"error": "Internal server error", "details": str(e)}), 500


def add_llm_aid(aid):
    # global llm_aid_app_key_map, llm_app_key_aid_map
    # import hashlib
    # if aid.id in llm_aid_app_key_map:
    #     llm_app_key = llm_aid_app_key_map[aid.id]
    # else:
    #     llm_app_key = str(int(time.time())+actual_port)
    #     llm_app_key = hashlib.sha256(llm_app_key.encode()).hexdigest()
    #     llm_aid_app_key_map[aid.id] = llm_app_key
    # llm_app_key_aid_map[llm_app_key] = aid
    return True


def add_llm_api_key(aid,llm_app_key):
    # global llm_aid_app_key_map, llm_app_key_aid_map
    # llm_aid_app_key_map[aid.id] = llm_app_key
    # llm_app_key_aid_map[llm_app_key] = aid
    return True
    

def get_base_url(aid,llm_aid):
    # global actual_port
    # # 获取实际分配的端口号
    # return "http://127.0.0.1:"+str(actual_port)+"/"+llm_aid
    pass

def get_llm_api_key(aid_str:str):
    # 获取实际分配的端口号
    pass
    # global llm_aid_app_key_map,llm_app_key_aid_map
    # if aid_str not in llm_aid_app_key_map:
    #     import secrets,hashlib
    #     llm_app_key = secrets.token_hex(16)
    #     llm_app_key = hashlib.sha256(llm_app_key.encode()).hexdigest()
    #     llm_aid_app_key_map[aid_str] = llm_app_key
    # return llm_aid_app_key_map[aid_str]

def llm_server_is_running():
    # global is_running
    # return is_running
    pass

def __run_server():
    # 端口设为0让系统自动分配
    pass
    # try:
    #     global actual_port
    #     if actual_port == 0:
    #         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    #         sock.bind(('', 0))
    #         actual_port = sock.getsockname()[1]
    #         sock.close()
    #         app.run(host='127.0.0.1', port = actual_port, debug=False)
    #     else:
    #         app.run(host='127.0.0.1', port = actual_port, debug=False)

    #     global is_running
    #     is_running = True
    # except Exception as e:
    #     is_running = False
    #     #print(f"Flask服务启动失败,请检查端口占用后，重启服务")

def run_server(debug:bool = False,port:int = 0,llm_aid_app_key_map_h = {},llm_app_key_aid_map_h = {}):
    # 创建并启动子线程运行Flask服务
    pass
    # app.logger.disabled = (not debug)
    # global actual_port,llm_aid_app_key_map,llm_app_key_aid_map
    # try:
    #     actual_port = int(port)
    # except (ValueError, TypeError):
    #     actual_port = 0  # 如果转换失败，设置为默认值0
    # llm_aid_app_key_map = llm_aid_app_key_map_h
    # llm_app_key_aid_map = llm_app_key_aid_map_h
    # server_thread = threading.Thread(target=__run_server)
    # server_thread.daemon = True  # 设置为守护线程，主线程退出时会自动结束
    # server_thread.start()
    # 主线程可以继续执行其他任务

# 添加一个关闭服务器的路由
# @app.route('/shutdown', methods=['POST'])
def shutdown_server():
    pass
    # func = request.environ.get('werkzeug.server.shutdown')
    # if func is None:
    #     return jsonify({"status": "error", "message": "Not running with Werkzeug Server"})
    # func()
    # return jsonify({"status": "success", "message": "Server shutting down..."})


# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from typing import Dict, List, Union, Iterable, Optional
from openai._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from openai.types.shared.chat_model import ChatModel
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.shared_params.metadata import Metadata
from openai.types.shared.reasoning_effort import ReasoningEffort
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_deleted import ChatCompletionDeleted
from openai.types.chat.chat_completion_tool_param import ChatCompletionToolParam
from openai.types.chat.chat_completion_audio_param import ChatCompletionAudioParam
from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam
from openai.types.chat.chat_completion_stream_options_param import ChatCompletionStreamOptionsParam
from openai.types.chat.chat_completion_prediction_content_param import ChatCompletionPredictionContentParam
from openai.types.chat.chat_completion_tool_choice_option_param import ChatCompletionToolChoiceOptionParam
from openai.types.chat import (
    ChatCompletionAudioParam,
    completion_list_params,
    completion_create_params,
    completion_update_params,
)
from typing_extensions import Literal, overload

class OpenAIMessage:    
    def __init__(self):
        self.messages :  Iterable[ChatCompletionMessageParam] = None
        self.model: Union[str, ChatModel] = None
        self.audio: Optional[ChatCompletionAudioParam] = None
        self.frequency_penalty: Optional[float] = None
        self.function_call: completion_create_params.FunctionCall = None
        self.functions: Iterable[completion_create_params.Function] = None
        self.logit_bias: Optional[Dict[str, int]] = None
        self.logprobs: Optional[bool] = None
        self.max_completion_tokens: Optional[int] = None
        self.max_tokens: Optional[int] = None
        self.metadata: Optional[Metadata] = None
        self.modalities: Optional[List[Literal["text", "audio"]]] = None
        self.n: Optional[int] = None
        self.parallel_tool_calls = None
        self.prediction: Optional[ChatCompletionPredictionContentParam] = None
        self.presence_penalty: Optional[float] = None
        self.reasoning_effort: Optional[ReasoningEffort] = None
        self.response_format: completion_create_params.ResponseFormat  = None
        self.seed: Optional[int]  = None
        self.service_tier: Optional[Literal["auto", "default"]]  = None
        self.stop: Union[Optional[str], List[str], None] = None
        self.store: Optional[bool] = None
        self.stream: Optional[Literal[False]] = None
        self.stream_options: Optional[ChatCompletionStreamOptionsParam]  = None
        self.temperature: Optional[float] = None
        self.tool_choice: ChatCompletionToolChoiceOptionParam = None
        self.tools: Iterable[ChatCompletionToolParam] = None
        self.top_logprobs: Optional[int]  = None
        self.top_p: Optional[float]  = None
        self.user: str  = None
        self.web_search_options: completion_create_params.WebSearchOptions = None
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        self.extra_headers: Headers  = None
        self.extra_query: Query  = None
        self.extra_body: Body  = None
        self.timeout: float  = 600  # 10分钟超时
    
    def to_json(self) -> str:
        """将对象序列化为JSON字符串，自动跳过None值"""
        data = {
            key: value 
            for key, value in self.__dict__.items()
            if value is not None and not key.startswith('_')
        }
        return json.dumps(data, 
                         default=lambda o: o.__dict__, 
                         ensure_ascii=False, 
                         separators=(',', ':'))


# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Union
import uuid
import time
import requests
from cryptography import x509
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.x509.oid import NameOID
from agentcp.ca.ca_root import CARoot
from agentcp.base.log import log_error, log_exception, log_info, log_warning
from agentcp.utils import tls_config as tls
from cryptography.hazmat.backends import default_backend
import os


class AuthClient:

    success_crt_list = []

    # HTTP 请求超时配置 (连接超时, 读取超时)
    HTTP_TIMEOUT = (3, 10)

    def __init__(self, agent_id: str, server_url: str, aid_path: str, seed_password: str):
        """认证客户端类
        Args:
            agent_id: 代理ID
            server_url: 服务器URL
        """
        self.agent_id = agent_id
        self.server_url = server_url
        self.signature = None
        self.aid_path = aid_path
        self.seed_password = seed_password

    def sign_in(self, max_retry_num: int = 10) -> Union[dict, None]:
        """登录方法，使用循环重试，失败返回 None"""
        for retry_count in range(max_retry_num + 1):
            try:
                if retry_count > 0:
                    # 指数退避：2s, 4s, 6s, ..., 最大 30s
                    backoff = min(2 * retry_count, 30)
                    log_info(f"Sign in retry {retry_count}/{max_retry_num}, waiting {backoff}s...")
                    time.sleep(backoff)

                hb_url = self.server_url + "/sign_in"
                log_info(f"Sign in: {hb_url}")
                request_id = uuid.uuid4().hex
                data = {
                    "agent_id": self.agent_id,
                    "request_id": request_id,
                }
                headers = {
                    'User-Agent': f'AgentCP/{__import__("agentcp").__version__} (AuthClient; {self.agent_id})'
                }
                response = requests.post(hb_url, json=data, verify=tls.requests_verify(), headers=headers, proxies={}, timeout=self.HTTP_TIMEOUT)

                if response.status_code == 200:
                    log_info(f"Sign in url: {hb_url}, response: {response.json()}")
                    aid_path = os.path.join(self.aid_path, self.agent_id + ".key")
                    private_key = self.__load_private_key(aid_path)
                    if private_key is None:
                        raise Exception("私钥加载失败,请检查加密种子是否一致")

                    aid_path = os.path.join(self.aid_path, self.agent_id + ".crt")
                    with open(aid_path, "rb") as f:
                        certificate_pem = f.read().decode('utf-8')

                    cert = x509.load_pem_x509_certificate(certificate_pem.encode('utf-8'))
                    public_key = cert.public_key()
                    public_key_pem = public_key.public_bytes(
                        encoding=serialization.Encoding.PEM,
                        format=serialization.PublicFormat.SubjectPublicKeyInfo
                    ).decode('utf-8')

                    if "nonce" in response.json():
                        nonce = response.json()["nonce"]
                        server_cert_valid = False

                        if "cert" not in response.json() or "signature" not in response.json():
                            raise Exception("服务器响应缺少 cert 或 signature 字段，无法验证服务器身份")

                        try:
                            cert_data = response.json()["cert"].encode('utf-8')
                            server_cert = x509.load_pem_x509_certificate(cert_data, default_backend())
                            server_public_key = server_cert.public_key()
                            data_to_verify = (self.agent_id + str(request_id)).lower().encode('utf-8')
                            signature_bytes = bytes.fromhex(response.json()["signature"])
                            server_public_key.verify(
                                signature_bytes,
                                data_to_verify,
                                ec.ECDSA(hashes.SHA256())
                            )
                            log_info("服务器签名验证成功")
                            server_cert_valid = self.__check_server_cert(server_cert)
                            if not server_cert_valid:
                                raise Exception("服务器证书验证失败")
                        except Exception as e:
                            server_cert_valid = False
                            raise Exception(f"{e}")

                        if not server_cert_valid:
                            raise Exception("服务器证书是无效证书，请注意通信安全")

                        if nonce:
                            signature = private_key.sign(
                                nonce.encode('utf-8'),
                                ec.ECDSA(hashes.SHA256())
                            )
                            data = {
                                "agent_id": self.agent_id,
                                "request_id": request_id,
                                "nonce": nonce,
                                "public_key": public_key_pem,
                                "cert": certificate_pem,
                                "signature": signature.hex(),
                            }
                            response = requests.post(hb_url, json=data, verify=tls.requests_verify(), headers=headers, proxies={}, timeout=self.HTTP_TIMEOUT)
                            if response.status_code == 200:
                                result = response.json()
                                self.signature = result.get("signature")
                                log_info("Sign in successful")
                                return result
                            else:
                                log_error(f"Sign in FAILED: {response.status_code} - {response.json().get('error', '')}")
                else:
                    log_error(f"Sign in failed: {response.status_code} - {response.json().get('error', '')}")

            except Exception as e:
                log_warning(f"Sign in exception (retry {retry_count}/{max_retry_num}): {e}")

        # 所有重试都失败
        log_error(f"Sign in failed after {max_retry_num} retries, please check network connection")
        return None
            
            
    def __load_private_key(self, aid_path):
        try:
            with open(aid_path, "rb") as f:
                private_key = serialization.load_pem_private_key(
                    f.read(),
                    password=self.seed_password.encode('utf-8'),
                )
            return private_key
        except Exception as e:
            # 兼容性代码，按照不加密获取 private_key
            return None

    def sign_out(self) -> None:
        """登出方法"""
        try:
            if self.signature is None:
                return
            hb_url = self.server_url + "/sign_out"
            data = {
                "agent_id": self.agent_id,
                "signature": self.signature,
            }
            headers = {
                'User-Agent': f'AgentCP/{__import__("agentcp").__version__} (AuthClient; {self.agent_id})'
            }
            response = requests.post(hb_url, json=data, verify=tls.requests_verify(), headers=headers, proxies={}, timeout=self.HTTP_TIMEOUT)
            if response.status_code == 200:
                log_info(f"Sign out OK: {response.json()}")
            else:
                log_error(f"Sign out failed: {response.json()}")
        except Exception as e:
            log_exception("Sign out exception")
            
    def __check_server_cert(self, server_cert):
        try:
            # 尝试获取主体的组织及 common name
            try:
                organization = server_cert.subject.get_attributes_for_oid(NameOID.ORGANIZATION_NAME)
                common_name = server_cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)
                if organization:
                    log_info(f"__check_server_cert 证书的组织名称为: {organization[0].value}")
                else:
                    log_info("__check_server_cert 未找到证书的组织名称")
                if common_name:
                    log_info(f"__check_server_cert 证书的通用名称为: {common_name[0].value}")
                else:
                    log_info("__check_server_cert 未找到证书的通用名称")
            except AttributeError:
                log_error("获取证书主体信息时出错，可能证书对象不存在或格式不正确。")
                return False
            aia_ext = server_cert.extensions.get_extension_for_class(x509.AuthorityInformationAccess)
            log_info(f"aia_ext:{aia_ext}")
            issuer_url = None
            for desc in aia_ext.value:
                if desc.access_method == x509.oid.AuthorityInformationAccessOID.CA_ISSUERS:
                    issuer_url = desc.access_location.value
                    break
    
            if issuer_url:
                log_info(f"证书颁发者 URL: {issuer_url}")
                if issuer_url in AuthClient.success_crt_list:
                    log_info(f"证书之前验证成功 {issuer_url}")
                    return True
                try:
                    issuer_response = requests.get(issuer_url, verify=tls.requests_verify(), proxies={}, timeout=self.HTTP_TIMEOUT)
                    issuer_response.raise_for_status()
                    #TODO:将证书内存以issuer_url为键值缓存在本地，避免重复下载和验证证书
                    issuer_cert = x509.load_pem_x509_certificate(issuer_response.content, default_backend())
                    # 验证服务器证书的有效性
                    issuer_public_key = issuer_cert.public_key()
                    issuer_public_key.verify(
                        server_cert.signature,
                        server_cert.tbs_certificate_bytes,
                        ec.ECDSA(server_cert.signature_hash_algorithm)
                    )
                    #TODO标记证书是受信任的
                    log_info(f"证书验证成功 {issuer_url}")
                    AuthClient.success_crt_list.append(issuer_url)
                    return self.__check_server_cert(issuer_cert)
                except requests.RequestException as e:
                    log_error(f"下载证书颁发者证书时出错: {e}")
                    return False
                except Exception as e:
                    log_error(f"验证证书时出错: {e}")
                    return False
            else:
                log_error("未找到证书颁发者信息")
                return False
        except x509.ExtensionNotFound:
            log_error("证书中未包含 AIA 扩展信息")
            try:
                root_cert_pem = CARoot().get_ca_root_crt()
                root_cert = x509.load_pem_x509_certificate(root_cert_pem.encode('utf-8'), default_backend())
                root_public_key = root_cert.public_key()
                root_public_key.verify(
                    server_cert.signature,
                    server_cert.tbs_certificate_bytes,
                    ec.ECDSA(server_cert.signature_hash_algorithm)
                )
                log_info("用root.crt 证书进行验证成功")
                return True
            except FileNotFoundError:
                log_error("未找到 root.crt 文件，请检查文件路径。")
                return False
            except Exception as e:
                log_error(f"加载 root.crt 证书时出错: {e}")
                return False
        except Exception as e:
            log_error(f"服务器证书验证失败: {e}")
            return False



# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import queue
import threading
from typing import Callable

from .exceptions import SDKError


class AtomicErrorContext:
    """
    错误收集，只有在订阅后才运行，否则不会有任何错误信息加入到队列
    """

    def __init__(self, max_queue_size=1024):
        self.queue = queue.Queue[SDKError](maxsize=max_queue_size)
        self.stop_flag = threading.Event()
        self.start_flag = threading.Event()
        self.worker_thread = threading.Thread(target=self._worker, daemon=True)

    def publish(self, e: SDKError):
        if not self.start_flag.is_set():
            return

        try:
            if self.queue.full():
                self.queue.get_nowait()
            e.agent_id = self.aid

            self.queue.put(e, block=False)
        except queue.Full:
            pass

    def subscribe(self, aid: str, func: Callable[[SDKError], None]):
        """
        上层直接退出登录线程未完全退出，防止重复订阅（多账号来回切换会出现异常）
        """
        if self.start_flag.is_set():
            return

        self.aid = aid
        self.func = func
        self.start_flag.set()
        self.worker_thread.start()
        atexit.register(self.close)

    def _worker(self):
        """后台线程，定时或批量处理队列数据"""
        while not self.stop_flag.is_set():
            try:
                # 等待队列有数据
                data = self.queue.get(timeout=1)
                self.func(data)
            except queue.Empty:
                pass  # 到达interval时间也可能队列为空

    def close(self):
        """关闭采集线程，确保队列数据全部发送"""
        self.stop_flag.set()
        self.worker_thread.join()



# -*- coding: utf-8 -*-
import os
from contextlib import contextmanager
from urllib.parse import urlparse


_LOCAL_HOSTS = ("localhost", "127.0.0.1", "::1", "0.0.0.0")
_LOCAL_NO_PROXY = "localhost,127.0.0.1,::1"
LOCAL_NO_PROXY_LIST = ["localhost", "127.0.0.1", "::1"]


def ensure_no_proxy_for_local_env() -> None:
    """Ensure NO_PROXY/no_proxy includes localhost, as a safety net against global proxies."""
    for key in ("NO_PROXY", "no_proxy"):
        existing = (os.environ.get(key) or "").strip()
        if not existing:
            os.environ[key] = _LOCAL_NO_PROXY
            continue

        existing_parts = [p.strip() for p in existing.replace(";", ",").split(",") if p.strip()]
        existing_lower = {p.lower() for p in existing_parts}
        for host in _LOCAL_NO_PROXY.split(","):
            if host.lower() not in existing_lower:
                existing_parts.append(host)
        os.environ[key] = ",".join(existing_parts)


def is_local_url(url: str) -> bool:
    """Return True when URL hostname is localhost/loopback."""
    try:
        parsed = urlparse(url)
        host = parsed.hostname
        if not host:
            return False
        host = host.strip("[]").lower()
        return host in _LOCAL_HOSTS
    except Exception:
        return False


def get_requests_proxies(use_system_proxy: bool, url: str):
    """
    For requests:
      - {} disables proxy
      - None uses env/system proxy
    """
    return {}


def get_trust_env(use_system_proxy: bool, url: str) -> bool:
    """For httpx/aiohttp trust_env: never trust env proxy for any URL."""
    return False


def pop_proxy_env():
    """Remove proxy env vars and return the saved mapping for restoration."""
    keys = (
        "HTTP_PROXY",
        "HTTPS_PROXY",
        "ALL_PROXY",
        "FTP_PROXY",
        "SOCKS_PROXY",
        "http_proxy",
        "https_proxy",
        "all_proxy",
        "ftp_proxy",
        "socks_proxy",
    )
    saved = {k: os.environ.get(k) for k in keys}
    for k in keys:
        os.environ.pop(k, None)
    return saved


def restore_proxy_env(saved) -> None:
    if not saved:
        return
    for k, v in saved.items():
        if v is None:
            os.environ.pop(k, None)
        else:
            os.environ[k] = v


@contextmanager
def without_proxy_env(enabled: bool = True):
    """
    Temporarily remove proxy env vars (HTTP(S)_PROXY/ALL_PROXY and lowercase variants).
    Useful for websocket libraries that consult env proxies.
    """
    if not enabled:
        yield
        return

    saved = pop_proxy_env()
    try:
        yield
    finally:
        restore_proxy_env(saved)



# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
集中式 TLS/SSL 配置模块

所有 TLS 决策（是否校验、用哪个 CA、是否禁用告警）只在此模块中做，
调用方只需使用以下 API：

- init_tls(app_path, debug, ca_bundle, verify)  初始化（通常在 AgentID.__init__ 中调用一次）
- requests_verify()          -> bool | str       给 requests 库用
- ssl_context()              -> ssl.SSLContext    给 websockets / urllib 用
- websocket_sslopt()         -> dict             给 websocket-client 用
- aiohttp_ssl()              -> SSLContext | bool 给 aiohttp 用
- allow_insecure()           -> bool             查询当前是否处于不安全模式
- ca_bundle_path()           -> str | None       获取 CA bundle 路径
- install_warning_filters()                      仅在开发模式+显式允许时禁用 InsecureRequestWarning

配置优先级：显式参数 > 环境变量 > 默认值

环境变量：
- AGENTCP_TLS_VERIFY           是否验证证书 (默认 true)
- AGENTCP_TLS_ALLOW_INSECURE   是否允许不安全模式 (默认 false)
- AGENTCP_TLS_CA_BUNDLE        自定义 CA bundle 路径
- AGENTCP_TLS_DISABLE_WARNINGS 是否禁用 InsecureRequestWarning (默认 false)
"""

import os
import ssl
import threading
from dataclasses import dataclass
from typing import Optional, Union

from agentcp.base.log import log_info, log_warning, log_error


# ==================== 内部状态 ====================

# 使用 RLock 以支持同一线程内的嵌套调用（如 ssl_context -> _state -> init_tls）
_STATE_LOCK = threading.RLock()
_STATE: Optional["TLSState"] = None
_SSL_CTX_CACHE: Optional[ssl.SSLContext] = None
_WS_SSLOPT_CACHE: Optional[dict] = None


@dataclass
class TLSState:
    """TLS 配置状态"""
    verify: bool
    allow_insecure: bool
    ca_bundle: Optional[str]
    debug: bool


# ==================== 工具函数 ====================

def _env_bool(name: str, default: bool) -> bool:
    """从环境变量读取布尔值"""
    val = os.getenv(name)
    if val is None:
        return default
    return val.strip().lower() in ("1", "true", "yes", "on")


def _strip_bom(text: str) -> str:
    """移除 UTF-8 BOM（Windows 记事本编辑过的文件常见）"""
    if text.startswith('\ufeff'):
        return text[1:]
    return text


def _build_ca_bundle(app_path: Optional[str], override_path: Optional[str]) -> Optional[str]:
    """构建 CA bundle 文件

    将内置根证书 + Certs/root 目录下的 .crt 文件合并为一个 PEM bundle。
    注意：只读取 CARoot 内置证书，不调用 set_ca_root_crt，避免覆盖单例状态。

    Args:
        app_path: 应用数据目录
        override_path: 用户指定的 CA bundle 路径（优先使用）

    Returns:
        bundle 文件路径，或 None
    """
    if override_path:
        if os.path.isfile(override_path):
            return override_path
        log_warning(f"[TLS] CA bundle path does not exist or is not a file: {override_path}, falling back to default")

    if not app_path:
        return None

    root_dir = os.path.join(app_path, "Certs", "root")
    os.makedirs(root_dir, exist_ok=True)
    bundle = os.path.join(root_dir, "ca-bundle.pem")

    parts = []

    # 收集系统 CA 证书（certifi 是 requests 的依赖，一定存在）
    try:
        import certifi
        with open(certifi.where(), "r", encoding="utf-8") as f:
            system_cas = _strip_bom(f.read()).strip()
            if system_cas:
                parts.append(system_cas)
    except Exception as e:
        log_warning(f"[TLS] 加载系统 CA 证书失败: {e}")

    # 收集内置根证书（只读取内置列表，不修改 CARoot 单例的 __ca_root_path）
    try:
        from agentcp.ca.ca_root import CARoot
        ca = CARoot()
        # 只读取内置证书数量（__ca_crt 列表），不调用 set_ca_root_crt
        for i in range(len(ca._CARoot__ca_crt)):
            pem = ca._CARoot__ca_crt[i]
            if pem:
                parts.append(_strip_bom(pem).strip())
    except Exception as e:
        log_warning(f"[TLS] 加载内置根证书失败: {e}")

    # 收集目录内 .crt 文件（排除 bundle 自身）
    try:
        for name in sorted(os.listdir(root_dir)):
            if name.endswith(".crt") and name != "ca-bundle.pem":
                filepath = os.path.join(root_dir, name)
                with open(filepath, "r", encoding="utf-8") as f:
                    content = _strip_bom(f.read()).strip()
                    if content and content not in parts:
                        parts.append(content)
    except Exception as e:
        log_warning(f"[TLS] 扫描证书目录失败: {e}")

    if parts:
        try:
            with open(bundle, "w", encoding="utf-8", newline="\n") as f:
                f.write("\n".join(parts) + "\n")
            return bundle
        except Exception as e:
            log_error(f"[TLS] 写入 CA bundle 失败: {e}")

    return None


# ==================== 公开 API ====================

def init_tls(
    app_path: Optional[str] = None,
    debug: bool = False,
    ca_bundle: Optional[str] = None,
    verify: Optional[bool] = None,
) -> None:
    """初始化 TLS 配置（通常只需调用一次）

    Args:
        app_path: 应用数据目录（用于定位 Certs/root）
        debug: 是否为调试模式
        ca_bundle: 自定义 CA bundle 路径
        verify: 是否验证证书（None 表示从环境变量读取）
    """
    global _STATE, _SSL_CTX_CACHE, _WS_SSLOPT_CACHE
    with _STATE_LOCK:
        _SSL_CTX_CACHE = None
        _WS_SSLOPT_CACHE = None

        insecure = _env_bool("AGENTCP_TLS_ALLOW_INSECURE", False)
        if verify is None:
            verify = _env_bool("AGENTCP_TLS_VERIFY", True)

        bundle_override = ca_bundle or os.getenv("AGENTCP_TLS_CA_BUNDLE")
        bundle = _build_ca_bundle(app_path, bundle_override)

        _STATE = TLSState(
            verify=bool(verify),
            allow_insecure=insecure,
            ca_bundle=bundle,
            debug=debug,
        )

        # 初始化时打印一次模式信息
        if not verify and insecure:
            log_warning("[TLS] Insecure mode enabled: certificate verification is DISABLED")
        elif not verify and not insecure:
            log_warning(
                "[TLS] AGENTCP_TLS_VERIFY is disabled, but AGENTCP_TLS_ALLOW_INSECURE is not set. "
                "Certificate verification remains ENABLED for safety. "
                "Set AGENTCP_TLS_ALLOW_INSECURE=1 to explicitly allow insecure mode."
            )
        elif bundle:
            log_info(f"[TLS] Certificate verification enabled, CA bundle: {bundle}")
        else:
            log_info("[TLS] Certificate verification enabled, using system CA store")


def _state() -> TLSState:
    """获取当前 TLS 状态（线程安全，double-check locking）

    使用 RLock 保证 ssl_context() -> _state() -> init_tls() 嵌套调用不会死锁。
    """
    global _STATE
    if _STATE is None:
        with _STATE_LOCK:
            if _STATE is None:
                init_tls()
    return _STATE


def allow_insecure() -> bool:
    """当前是否处于不安全模式（verify=False 且 allow_insecure=True）"""
    s = _state()
    return (not s.verify) and s.allow_insecure


def ca_bundle_path() -> Optional[str]:
    """获取 CA bundle 路径"""
    return _state().ca_bundle


def requests_verify() -> Union[bool, str]:
    """获取 requests 库的 verify 参数

    Returns:
        - False: 不安全模式，跳过验证
        - str: CA bundle 路径
        - True: 使用系统默认 CA
    """
    s = _state()
    if (not s.verify) and s.allow_insecure:
        return False
    return s.ca_bundle if s.ca_bundle else True


def ssl_context() -> ssl.SSLContext:
    """获取 SSLContext（带缓存）

    用于 websockets 库和 urllib。
    """
    global _SSL_CTX_CACHE
    if _SSL_CTX_CACHE is not None:
        return _SSL_CTX_CACHE

    with _STATE_LOCK:
        if _SSL_CTX_CACHE is not None:
            return _SSL_CTX_CACHE

        s = _state()
        if (not s.verify) and s.allow_insecure:
            ctx = ssl.create_default_context()
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        elif s.ca_bundle:
            ctx = ssl.create_default_context(cafile=s.ca_bundle)
        else:
            ctx = ssl.create_default_context()

        _SSL_CTX_CACHE = ctx
        return ctx


def aiohttp_ssl() -> Union[ssl.SSLContext, bool]:
    """获取 aiohttp 的 ssl 参数

    Returns:
        - False: 不安全模式
        - SSLContext: 正常验证
    """
    if allow_insecure():
        return False
    return ssl_context()


def websocket_sslopt() -> dict:
    """获取 websocket-client 库的 sslopt 参数（带缓存）"""
    global _WS_SSLOPT_CACHE
    if _WS_SSLOPT_CACHE is not None:
        return _WS_SSLOPT_CACHE

    with _STATE_LOCK:
        if _WS_SSLOPT_CACHE is not None:
            return _WS_SSLOPT_CACHE

        if allow_insecure():
            opt = {
                "cert_reqs": ssl.CERT_NONE,
                "check_hostname": False,
            }
        else:
            ca = ca_bundle_path()
            if ca:
                opt = {
                    "cert_reqs": ssl.CERT_REQUIRED,
                    "check_hostname": True,
                    "ca_certs": ca,
                }
            else:
                opt = {
                    "cert_reqs": ssl.CERT_REQUIRED,
                    "check_hostname": True,
                }

        _WS_SSLOPT_CACHE = opt
        return opt


def install_warning_filters() -> None:
    """在不安全模式下禁用 InsecureRequestWarning

    仅当 allow_insecure=True 且 AGENTCP_TLS_DISABLE_WARNINGS=True 时生效。
    """
    if allow_insecure() and _env_bool("AGENTCP_TLS_DISABLE_WARNINGS", False):
        try:
            import urllib3
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            log_warning("[TLS] InsecureRequestWarning has been suppressed")
        except Exception:
            pass



# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from agentcp.base.log import log_error, log_info


class Mermaid:
    def __init__(self, content):
        self.mermaid_code = content
        self.graph_type = None
        self.graph_direction = None  # 新增：存储图的方向
        self.nodes = []
        self.node_dict = {}
        self.node_styles = {}
        self.edges = []
        self.parse_mermaid()

        log_info(f"图类型:{self.graph_type}")
        log_info(f"总共{len(self.nodes)}个节点,节点列表如下:\n\t{self.nodes}")
        for node in self.node_dict.items():
            log_info(f"节点映射:{node}")
        for edge in self.edges:
            log_info(f"边关系:{edge}")


    @classmethod
    def from_file(cls, dir_path, file_name):
        return cls(cls.read_mermaid_code(dir_path, file_name))
    def read_mermaid_code(dir_path: str, file_name: str):
        file_path = Path(dir_path) / f"{file_name}.mmd"
        absolute_path = file_path.absolute()
        log_info(f"读取Mermaid文件，绝对路径: {absolute_path}")
        if file_path.is_file():
            try:
                with file_path.open('r', encoding='utf-8') as f:
                    content = f.read()
                    content = content.replace('&nbsp;', '').replace(u'\xa0', u'')
                    return content
            except Exception as e:
                log_error(f"读取文件错误: {e}")
                return ""
        else:
            log_error(f"文件 {file_path} 未找到")
            return ""

    def parse_mermaid(self):
        if not self.mermaid_code:
            return

        # 检查图类型和方向
        for line in self.mermaid_code.split('\n'):
            line = line.strip()
            if not line or line.startswith('%'):
                continue

            if line.startswith('graph'):
                parts = line.split()
                if len(parts) >= 2:
                    self.graph_type = 'graph'
                    self.graph_direction = parts[1]  # 存储方向 (TD, LR, BT, RL等)
                break
            elif line.startswith(('pie', 'sequenceDiagram', 'gantt', 'classDiagram', 'stateDiagram')):
                self.graph_type = line.split()[0]
                log_error(f"不支持 {self.graph_type} 类型的图")
                return
            else:
                log_error("无法识别的图类型")
                return

        if self.graph_type != 'graph':
            return

        # 解析内容
        for line in self.mermaid_code.split('\n'):
            line = line.strip()
            if not line or line.startswith(('%', 'classDef', 'linkStyle', 'style', 'graph')):
                continue

            # 解析节点
            if (any(x in line for x in ['[', '(', '{']) and
                not any(x in line for x in ['-->', '--', '->'])) or \
                    (not any(x in line for x in ['-->', '--', '->']) and
                     line not in self.node_dict and
                     not line.startswith('graph')):
                self.parse_node(line)

            # 解析边关系
            if any(x in line for x in ['-->', '--', '->']):
                self.parse_edge(line)

    def parse_node(self, line):
        """解析节点定义"""
        node_name = ''
        node_desc = ''
        style_class = ''

        # 提取样式类 (:::xxx)
        if ':::' in line:
            parts = line.split(':::')
            line = parts[0].strip()
            style_class = parts[1].strip()

        # 花括号格式: A{描述}
        if '{' in line and '}' in line:
            start = line.find('{')
            end = line.find('}')
            node_name = line[:start].strip()
            node_desc = line[start + 1:end].strip('" ')

        # 方括号格式: A["描述"]
        elif '[' in line and ']' in line:
            start = line.find('[')
            end = line.find(']')
            node_name = line[:start].strip()
            node_desc = line[start + 1:end].strip('" ')

        # 圆括号格式: A(描述)
        elif '(' in line and ')' in line:
            start = line.find('(')
            end = line.find(')')
            node_name = line[:start].strip()
            node_desc = line[start + 1:end].strip('" ')

        # 无描述节点: A
        else:
            node_name = line.strip()
            node_desc = node_name

        if node_name and node_name not in self.node_dict:
            self.node_dict[node_name] = node_desc or node_name
            self.nodes.append(self.node_dict[node_name])
            if style_class:
                self.node_styles[node_name] = style_class

    def parse_edge(self, line):
        """解析边关系"""
        line = line.replace(' ', '')

        # 处理 A-->|描述|B 格式
        if '-->|' in line and '|' in line[line.find('-->|') + 3:]:
            arrow_pos = line.find('-->|')
            desc_start = arrow_pos + 4
            desc_end = line.find('|', desc_start)
            source = line[:arrow_pos]
            target = line[desc_end + 1:]
            description = line[desc_start:desc_end]
            self.edges.append((source, description, target))
            return

        # 处理 A-->B 格式
        if '-->' in line:
            parts = line.split('-->')
            if len(parts) == 2:
                self.edges.append((parts[0], '', parts[1]))
            return

        # 处理 A--描述-->B 格式
        if '--' in line and '-->' in line:
            arrow_pos = line.find('-->')
            source = line[:line.find('--')]
            target = line[arrow_pos + 3:]
            description = line[line.find('--') + 2:arrow_pos]
            self.edges.append((source, description, target))
            return

        # 处理 A->B 格式
        if '->' in line:
            parts = line.split('->')
            if len(parts) == 2:
                self.edges.append((parts[0], '', parts[1]))
            return
# 示例使用
if __name__ == "__main__":
    #mermaid = Mermaid.from_file('.', 'workflow')
    mmd = """
    graph TD
    %% ===== 节点定义 =====
    Z[用户交互层User]:::user
    A[任务拆解/推进/退出]:::Planner 
    B(个人助手):::PA
    D[Agent选择 AgentSelector]:::selector
    F[协作Agents]:::action
    G[单步任务执行]:::process
    Z -->|自然语言任务| B
    A -->|任务推进| D
    B -->|简单任务直接执行交付或复杂任务执行完成交付| Z
    B -->|复杂任务| A
    A -->|复杂任务完成所有步骤或者达成退出条件,交付| B
    D -->|任务推进|G
    G -->|调用| F
    F -->|成功/失败| G
    G -->|有失败任务未达到交付条件,重选Agent| D
    G -->|执行成功后交付|A
    """

    mermaid2 = Mermaid(mmd)



# -*- coding: utf-8 -*-
"""
滑动窗口指标统计

支持多时间窗口的实时统计分析
"""

import time
from typing import Dict, List, Tuple, Any


class TimeWindow:
    """单个时间窗口

    维护一个固定时长的滑动窗口，自动清理过期数据点
    """

    def __init__(self, duration_seconds: int):
        """初始化时间窗口

        Args:
            duration_seconds: 窗口时长（秒）
        """
        self.duration = duration_seconds
        self.data_points: List[Tuple[float, dict]] = []  # [(timestamp, metrics), ...]

    def add_snapshot(self, timestamp: float, metrics: dict):
        """添加一个时间点的指标快照

        Args:
            timestamp: 时间戳
            metrics: 指标字典，必须包含以下字段：
                - received_delta: 增量接收消息数
                - success_delta: 增量成功派发数
                - failed_delta: 增量失败数
                - avg_latency: 平均延迟（毫秒）
                - queue_size: 当前队列大小
        """
        self.data_points.append((timestamp, metrics))

        # 清理过期数据（保留窗口时长内的数据）
        cutoff = timestamp - self.duration
        self.data_points = [(t, m) for t, m in self.data_points if t >= cutoff]

    def get_stats(self) -> Dict[str, Any]:
        """计算窗口内的统计数据

        Returns:
            包含以下统计指标的字典：
            - throughput_per_second: 吞吐量（消息/秒）
            - avg_latency_ms: 平均延迟（毫秒）
            - success_rate: 成功率（%）
            - total_messages: 窗口内总消息数
            - failed_messages: 窗口内失败消息数
            - avg_queue_size: 平均队列大小
            - window_duration: 窗口时长（秒）
            - data_points_count: 数据点数量
        """
        if not self.data_points:
            return {
                'throughput_per_second': 0.0,
                'avg_latency_ms': 0.0,
                'success_rate': 0.0,
                'total_messages': 0,
                'failed_messages': 0,
                'avg_queue_size': 0,
                'window_duration': self.duration,
                'data_points_count': 0,
            }

        # 计算总量
        total_received = sum(m.get('received_delta', 0) for _, m in self.data_points)
        total_success = sum(m.get('success_delta', 0) for _, m in self.data_points)
        total_failed = sum(m.get('failed_delta', 0) for _, m in self.data_points)

        # 计算实际时间跨度（可能小于窗口时长）
        actual_duration = self.data_points[-1][0] - self.data_points[0][0]
        if actual_duration < 1:
            actual_duration = 1  # 避免除零

        # 计算吞吐量 (msg/s)
        throughput = total_received / actual_duration

        # 计算平均延迟（只考虑有延迟数据的点）
        latencies = [m.get('avg_latency', 0) for _, m in self.data_points if m.get('avg_latency', 0) > 0]
        avg_latency = sum(latencies) / len(latencies) if latencies else 0.0

        # 计算成功率
        success_rate = (total_success / max(total_received, 1)) * 100

        # 计算平均队列大小
        queue_sizes = [m.get('queue_size', 0) for _, m in self.data_points]
        avg_queue_size = sum(queue_sizes) / len(queue_sizes) if queue_sizes else 0

        return {
            'throughput_per_second': round(throughput, 2),
            'avg_latency_ms': round(avg_latency, 2),
            'success_rate': round(success_rate, 2),
            'total_messages': total_received,
            'failed_messages': total_failed,
            'avg_queue_size': round(avg_queue_size, 1),
            'window_duration': self.duration,
            'data_points_count': len(self.data_points),
        }


class SlidingWindowMetrics:
    """多时间窗口管理器

    管理多个不同时长的滑动窗口，提供多粒度的统计视图
    支持的窗口：1分钟、3分钟、5分钟、10分钟、15分钟
    """

    def __init__(self):
        """初始化多时间窗口管理器"""
        # 创建多个时间窗口
        self.windows: Dict[str, TimeWindow] = {
            '1m': TimeWindow(60),      # 1分钟
            '3m': TimeWindow(180),     # 3分钟
            '5m': TimeWindow(300),     # 5分钟
            '10m': TimeWindow(600),    # 10分钟
            '15m': TimeWindow(900),    # 15分钟
        }

        # 记录上一次快照的状态
        self.last_snapshot_time = 0
        self.last_metrics_snapshot: Dict[str, Any] = {}

    def update(self, current_metrics: dict):
        """更新所有窗口（应该每10秒调用一次）

        Args:
            current_metrics: 当前累计指标，必须包含以下字段：
                - received_total: 累计接收消息总数
                - dispatched_success: 累计派发成功数
                - dispatched_failed: 累计派发失败数
                - dispatch_queue_size: 当前派发队列大小
                - avg_dispatch_latency_ms: 平均派发延迟（可选）
                - avg_handler_latency_ms: 平均处理延迟（可选）
        """
        now = time.time()

        # 计算增量指标 (delta)
        metrics_delta = self._calculate_delta(current_metrics)

        # 更新所有窗口
        for window in self.windows.values():
            window.add_snapshot(now, metrics_delta)

        # 保存当前快照状态
        self.last_snapshot_time = now
        self.last_metrics_snapshot = current_metrics.copy()

    def _calculate_delta(self, current: dict) -> dict:
        """计算两次快照之间的增量

        Args:
            current: 当前累计指标

        Returns:
            增量指标字典
        """
        if not self.last_metrics_snapshot:
            # 第一次快照，返回零增量
            return {
                'received_delta': 0,
                'success_delta': 0,
                'failed_delta': 0,
                'avg_latency': 0.0,
                'queue_size': current.get('dispatch_queue_size', 0),
            }

        # 计算增量
        received_delta = current.get('received_total', 0) - self.last_metrics_snapshot.get('received_total', 0)
        success_delta = current.get('dispatched_success', 0) - self.last_metrics_snapshot.get('dispatched_success', 0)
        failed_delta = current.get('dispatched_failed', 0) - self.last_metrics_snapshot.get('dispatched_failed', 0)

        # 获取平均延迟（优先使用 dispatch_latency，其次使用 handler_latency）
        avg_latency = 0.0
        if 'avg_dispatch_latency_ms' in current:
            avg_latency = current['avg_dispatch_latency_ms']
        elif 'avg_handler_latency_ms' in current:
            avg_latency = current['avg_handler_latency_ms']

        return {
            'received_delta': max(0, received_delta),  # 防止负数
            'success_delta': max(0, success_delta),
            'failed_delta': max(0, failed_delta),
            'avg_latency': avg_latency,
            'queue_size': current.get('dispatch_queue_size', 0),
        }

    def get_window_stats(self, window_name: str) -> Dict[str, Any]:
        """获取指定窗口的统计数据

        Args:
            window_name: 窗口名称（'1m', '3m', '5m', '10m', '15m'）

        Returns:
            窗口统计数据字典，如果窗口不存在则返回空字典
        """
        if window_name not in self.windows:
            return {}
        return self.windows[window_name].get_stats()

    def get_all_windows(self) -> Dict[str, Dict[str, Any]]:
        """获取所有窗口的统计数据

        Returns:
            字典，key为窗口名称，value为统计数据
        """
        return {
            name: window.get_stats()
            for name, window in self.windows.items()
        }

    def reset(self):
        """重置所有窗口数据"""
        for window in self.windows.values():
            window.data_points.clear()
        self.last_snapshot_time = 0
        self.last_metrics_snapshot.clear()



# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import hashlib
import os

import requests

from agentcp.base.auth_client import AuthClient
from agentcp.base.client import IClient
from agentcp.base.log import log_debug, log_error, log_exception, log_info
from agentcp.utils import tls_config as tls

from ..context import ErrorContext, exceptions


class ApClient(IClient):
    def __init__(self, agent_id: str, server_url: str, aid_path: str, seed_password: str):
        super().__init__()
        self.agent_id = agent_id
        self.heartbeat_server = ""
        self.message_server = ""
        self.server_url = f"{server_url}" + "/api/accesspoint"
        self.auth_client = AuthClient(agent_id, self.server_url, aid_path, seed_password)  # 使用AuthClient

    def initialize(self):
        self.auth_client.sign_in()
        self.get_entrypoint_config()

    def sign_in(self) -> bool:
        return self.auth_client.sign_in() is not None

    def get_headers(self) -> dict:
        return {"User-Agent": f"AgentCP/{__import__('agentcp').__version__} (AuthClient; {self.agent_id})"}

    def sign_out(self):
        """登出方法"""
        self.auth_client.sign_out()

    def post_private_data(self, data):
        try:
            ep_url = self.server_url + "/post_agent_private_data"
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
                "data": data,
            }
            response = self.post_request(ep_url, json=data)
            if response.status_code == 200:
                log_info(f"post_private_data ok:{response.json()}")
                return response.json()["data"]
            else:
                log_error(f"post_private_data failed:{response.json()}")
                return None
        except Exception as e:
            log_exception(f"Post private data exception occurred: {e}")
            ErrorContext.publish(exceptions.SDKError(f"Post private data exception occurred: {e}"))
            return None

    def get_all_public_data(self):
        try:
            ep_url = self.server_url + "/get_all_public_data"
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
            }
            response = self.post_request(ep_url, json=data)
            if response.status_code == 200:
                return response.json()["data"]
            else:
                log_error(f"get_all_public_data failed:{response.json()}")
                return []
        except Exception as e:
            log_error("get_all_public_data exception:")
            log_exception(e)
            ErrorContext.publish(exceptions.SDKError(f"get_all_public_data exception: {e}"))
            return []

    def get_agent_list(self):
        try:
            ep_url = self.server_url + "/get_agent_list"
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
            }
            response = self.post_request(ep_url, json=data)
            if response.status_code == 200:
                log_info(f"get_all_public_data ok:{response.json()}")
                return response.json()["data"]
            else:
                log_error(f"get_all_public_data failed:{response.json()}")
                return None
        except Exception as e:
            log_exception(f"Get all public data exception occurred: {e}")
            ErrorContext.publish(exceptions.SDKError(f"Get all public data exception occurred: {e}"))

    def get_entrypoint_config(self):
        try:
            ep_url = self.server_url + "/get_accesspoint_config"
            log_debug(f"Get server config: {ep_url}")
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
            }
            response = self.post_request(ep_url, json=data)
            if response.status_code == 200:
                try:
                    config = response.json()
                    if isinstance(config.get("config"), str):  # 处理config是字符串的情况
                        import json

                        config["config"] = json.loads(config["config"])
                    if "config" in config:
                        if "heartbeat_server" in config["config"]:
                            self.heartbeat_server = config["config"]["heartbeat_server"]
                            log_debug(f"Set heartbeat server to: {self.heartbeat_server}")
                        if "message_server" in config["config"]:
                            self.message_server = config["config"]["message_server"]
                            log_debug(f"Set message server to: {self.message_server}")
                except (ValueError, AttributeError) as e:
                    ErrorContext.publish(exceptions.SDKError(f"处理ap返回的数据出错: {response.text}"))
                    log_exception(f"Failed to parse JSON. Response content: {response.text[:500]}. Error: {e}")
            else:
                log_error(f"Get entrypoint config {ep_url} failed:{response.json()}")

        except Exception as e:
            log_exception(f"Get entrypoint config exception occurred: {e}")
            ErrorContext.publish(exceptions.SDKError(f"Get entrypoint config exception occurred: {e}"))

    def get_agent_profile(self, agentid):
        try:
            url = f"https://{agentid}/agentprofile.json"
            response = self.get_request(url)
            if response.status_code == 200:
                log_debug(f"get_agent_public_data ok:{response.json()}")
                return response.json()
            else:
                log_error(f"get_agent_public_data failed:{response.json()}")
                return None
        except Exception as e:
            log_exception(f"Get agent public data exception occurred: {e}")
            return None

    def get_agent_public_data(self, agentid):
        try:
            ep_url = self.server_url + "/get_agent_public_data"
            data = {
                "agent_id": f"{agentid}",
                "signature": self.auth_client.signature,
            }
            response = self.post_request(ep_url, json=data)
            if response.status_code == 200:
                log_debug(f"get_agent_public_data ok:{response.json()}")
                return response.json()[
                    "data"
                ]  # Return the data if it exists, else return None or handle it as you like
            else:
                log_error(f"get_agent_public_data failed:{response.json()}")
                return None
        except Exception as e:
            log_exception(f"Get agent public data exception occurred: {e}")
            return None

    def get_agent_private_data(self):
        try:
            ep_url = self.server_url + "/get_agent_private_data"
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
            }
            response = self.post_request(ep_url, json=data)
            if response.status_code == 200:
                log_debug(f"Get_agent_private_data ok:{response.json()}")
                return response.json()["data"]
            else:
                log_error(f"Get_agent_private_data failed:{response.json()}")
                return None
        except Exception as e:
            log_exception(f"Get agent private data exception occurred: {e}")
            return None

    def get_heartbeat_server(self):
        return self.heartbeat_server

    def get_message_server(self):
        return self.message_server

    def __get_sha256(self, input_str: str) -> str:
        sha256_hash = hashlib.sha256()
        sha256_hash.update(input_str.encode("utf-8"))
        return sha256_hash.hexdigest()

    def __scan_folder(self, folder_path):
        file_list = []
        for root, dirs, files in os.walk(folder_path):
            for file in files:
                relative_file_path = os.path.relpath(os.path.join(root, file), folder_path)
                relative_file_path = relative_file_path.replace("\\", "/")
                file_path = os.path.join(root, file)

                try:
                    file_size = os.path.getsize(file_path)
                    last_modified_utc = datetime.datetime.fromtimestamp(
                        os.path.getmtime(file_path), tz=datetime.timezone.utc
                    )
                    last_modified_ms = int(last_modified_utc.timestamp() * 1000)
                    sha256 = self.__get_sha256(file_path)
                    file_info = {
                        "full_path": relative_file_path,
                        "size": file_size,
                        "last_modified": last_modified_ms,
                        "sha256": sha256,
                    }
                    file_list.append(file_info)
                except Exception as e:
                    import traceback

                    log_error(f"处理文件 {relative_file_path} 时出错，错误信息: {str(e)}")
                    log_error("详细堆栈信息如下：")
                    log_error(traceback.format_exc())
        return file_list

    def sync_public_files(self, file_path) -> bool:
        try:
            files_list = self.__scan_folder(file_path)
            ep_url = self.server_url + "/sync_public_files"
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
                "file_list": files_list,
            }
            response = requests.post(ep_url, json=data, verify=tls.requests_verify(), proxies={})
            if response.status_code == 200:
                rj = response.json()
                log_info(f"sync_public_files ok:{rj}")
                if "need_upload_files" in rj:
                    need_upload_files = rj["need_upload_files"]
                    for file_name in need_upload_files:
                        params = {
                            "agent_id": self.agent_id,
                            "signature": self.auth_client.signature,
                            "file_name": file_name,
                        }
                        full_path = os.path.join(file_path, file_name)
                        hb_url = self.server_url + "/upload_file"
                        with open(full_path, "rb") as file:
                            files = {"file": file}
                            response = requests.post(hb_url, data=params, files=files, verify=tls.requests_verify(), proxies={})
                            if response.status_code == 200:
                                log_info(f"文件{file_name}上传成功 => {response.json()['url']}")
                            else:
                                log_error(f"文件{file_name}上传失败，状态码: {response.status_code}")
                if "need_download_files" in rj:
                    need_download_files = rj["need_download_files"]
                    for file_name in need_download_files:
                        download_url = (
                            self.server_url
                            + "/download_file"
                            + f"?file_name={file_name}&agent_id={self.agent_id}&signature={self.auth_client.signature}"
                        )
                        log_info(f"download_url:{download_url}")
                        try:
                            # 发送 GET 请求下载文件
                            response = requests.get(download_url, verify=tls.requests_verify(), stream=True, proxies={})
                            response.raise_for_status()  # 检查请求是否成功
                            # 打开本地文件以写入二进制数据
                            save_path = os.path.join(file_path, file_name)
                            os.makedirs(os.path.dirname(save_path), exist_ok=True)
                            with open(save_path, "wb") as file:
                                # 逐块写入文件，避免一次性加载大文件到内存
                                for chunk in response.iter_content(chunk_size=16384):
                                    if chunk:  # 过滤掉保持活动的新块
                                        file.write(chunk)
                            log_info(f"文件下载成功，保存路径: {save_path}")
                        except requests.RequestException as e:
                            log_error(f"下载文件时发生请求错误: {e}")
                        except Exception as e:
                            log_error(f"下载文件时发生未知错误: {e}")
                return True
            else:
                raise Exception(f"sync_public_files failed:{response.status_code}")
                return False
        except Exception:
            import traceback

            log_error("详细堆栈信息如下：")
            log_error(traceback.format_exc())
            return False

    def delete_public_file(self, file_name: str):
        try:
            ep_url = self.server_url + "/delete_public_file"
            data = {
                "agent_id": f"{self.agent_id}",
                "signature": self.auth_client.signature,
                "file_name": file_name,
            }
            response = requests.post(ep_url, json=data, verify=tls.requests_verify(), proxies={})
            if response.status_code == 200:
                log_info(f"delete_public_file ok:{response.json()}")
            else:
                log_error(f"delete_public_file failed:{response.json()}")
        except Exception as e:
            log_exception(f"delete_public_file in exception: {e}")
        pass



# -*- coding: utf-8 -*-
# Copyright 2025 AgentUnion Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import platform as plt


class SDKError(Exception):
    trace_id = ""
    agent_id = ""
    platform = plt.version()
    py_version = plt.python_version()  # type: ignore

    def __init__(self, message, trace_id="",code = 1000):
        super().__init__(message)
        self.trace_id = trace_id
        import agentcp
        self.code = code
        self.sdk_version = f"agentcp/{agentcp.__version__}"

    def to_dict(self):
        return {
            "code": self.code,
            "trace_id": self.trace_id,
            "message": str(self),
            "agent_id": self.agent_id,
            "platform": self.platform,
            "py_version": self.py_version,
            "sdk_version": self.sdk_version,
        }


class LLMFlowError(SDKError):
    """调用 LLM 出错"""

    code = 1001


class CallLLMError(LLMFlowError):
    """调用 LLM 出错"""

    code = 1002


class CreateSessionError(LLMFlowError):
    """创建会话请求或响应出错"""

    code = 1003


class InviteMemberError(LLMFlowError):
    """邀请成员请求或响应出错"""

    code = 1005


class JoinSessionError(LLMFlowError):
    """加入会话请求或响应出错"""

    code = 1007


class SendMsgError(LLMFlowError):
    """发送消息出错"""

    code = 1009


class CreateStreamError(LLMFlowError):
    """创建流式通道请求或响应出错"""

    code = 1010


class StreamUrlNotifyError(LLMFlowError):
    """通知流地址出错"""

    code = 1012


class PullStreamError(LLMFlowError):
    """拉取流内容出错"""

    code = 1013


class SendChunkToStreamError(LLMFlowError):
    """向流中发送内容块出错"""

    code = 1014


class ChunksBySSEError(LLMFlowError):
    """通过 SSE 发送内容块出错"""

    code = 1015


class CloseStreamError(LLMFlowError):
    """关闭流式通道出错"""

    code = 1016



# -*- coding: utf-8 -*-
"""
AgentCp 监控模块

提供多时间窗口的监控统计、时间序列存储和实时监控服务
"""

from .sliding_window import SlidingWindowMetrics, TimeWindow
from .metrics_store import MetricsStore
from .monitoring_service import MonitoringService
from .standalone_reader import StandaloneMonitoringReader, get_standalone_reader

__all__ = [
    'SlidingWindowMetrics',
    'TimeWindow',
    'MetricsStore',
    'MonitoringService',
    'StandaloneMonitoringReader',
    'get_standalone_reader',
]
